{
   MegaBrutal's SMTP Server (MgSMTP)
   Copyright (C) 2010-2014 MegaBrutal

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
}

{
   Unit: Listener
   This unit is responsible for listening for incoming connections, and
   serve them, communicating by the SMTP protocol.

   It always places incoming e-mails in the spool, and lets it to process
   them later. However, this unit still links the Mailbox and Relay unit to
   verify addresses. The Policies unit also plays an important role, it
   determines what rights does the client have, and it authenticates users.
}


{$MODE DELPHI}
unit Listener;

interface
uses SysUtils, Classes, Base64, Network, NetRFC, RFCSMTP,
     Common, Log, Policies, Spool, Mailbox, Relay;

type

   TMgSMTPListener = class(TTCPListener)
      constructor Create(Port: word);
   protected
      procedure HandleClient(Connection: TTCPConnection); override;
      procedure ReceiveEMailData(TCP: TTCPRFCConnection; Response: TRFCReply; SpoolObject: TSpoolObjectCreator);
   end;


   procedure StartListeners;
   procedure StopListeners;


implementation

var

   MgSMTPListeners: array of TMgSMTPListener;


procedure StartListeners;
var i: integer;
begin
   SetLength(MgSMTPListeners, MainServerConfig.ListenPorts.Count);
   for i:= 0 to Length(MgSMTPListeners) - 1 do begin
      MgSMTPListeners[i]:= TMgSMTPListener.Create(StrToIntDef(MainServerConfig.ListenPorts.Strings[i], STANDARD_SMTP_PORT));
      MgSMTPListeners[i].StartListen;
   end;
end;

procedure StopListeners;
var i: integer;
begin
   for i:= 0 to Length(MgSMTPListeners) - 1 do begin
      MgSMTPListeners[i].StopListen;
      MgSMTPListeners[i].Free;
   end;
   SetLength(MgSMTPListeners, 0);
end;


function Base64Decode(Source: string): string;
var StringStream: TStringStream; Base64DecodingStream: TBase64DecodingStream;
    c: char;
begin
   StringStream:= TStringStream.Create(Source);
   Base64DecodingStream:= TBase64DecodingStream.Create(StringStream);
   Result:= '';
   while not Base64DecodingStream.EOF do begin
      Base64DecodingStream.Read(c, 1);
      Result:= Result + c;
   end;
   Base64DecodingStream.Destroy;
   StringStream.Destroy;
end;

procedure SetEMailProperties(Parameters: string; SpoolObject: TSpoolObject);
var CPrm, Rem, Key, Value: string;
begin
   { Cut down e-mail address. }
   SplitParameters(Parameters, CPrm, Rem);
   repeat
      SplitParameters(Rem, CPrm, Rem);
      SplitParameters(CPrm, Key, Value, '=');
      Key:= UpperCase(Key);
      if Key = 'SIZE' then SpoolObject.EMailProperties.Size:= StrToIntDef(Value, 0)
      else if Key = 'BODY' then begin
         if UpperCase(Value) = '8BITMIME' then
            SpoolObject.EMailProperties.SetFlag(EF_8BITMIME);
      end;
   until (Rem = '');
end;

function HandleRewrite(OriginalAddress: string; Mailbox: PMailbox; SpoolObject: TSpoolObjectCreator): string;
var i: integer;
begin
   for i:= 0 to Mailbox^.RewriteCount - 1 do
      SpoolObject.Envelope.AddRecipient(Mailbox^.GetRewriteToEntry(i));
   if Mailbox^.RewritePassThru then
      SpoolObject.Envelope.AddRecipient(OriginalAddress);
   if Mailbox^.RewriteCount > 0 then begin
      if Mailbox^.RewritePassThru then
         Result:= 'Rewrite: ' + OriginalAddress + ' -> ' + OriginalAddress + ',' + Mailbox^.GetRewriteToListStr
      else
         Result:= 'Rewrite: ' + OriginalAddress + ' -> ' + Mailbox^.GetRewriteToListStr;
   end
   else
      Result:= '';
end;


constructor TMgSMTPListener.Create(Port: word);
begin
   { Request connection objects with support for RFC-style commands & responses. }
   inherited Create(Port, NET_TCP_RFCSUPPORT);
   Logger.AddLine('Server', 'Listening on port: ' + IntToStr(Port));
end;


procedure TMgSMTPListener.HandleClient(Connection: TTCPConnection);
{ This is the procedure that actually handles the clients. It receives
  an object that manages the established connection in the parameter.
  TTCPConnection is defined in the Network unit. }
var
   TCP: TTCPRFCConnection;
   Originator: TIPNamePair;
   Response: TRFCReply;
   PolicyObject: TPolicyObject;
   SpoolObject: TSpoolObjectCreator;
   Cmd: shortstring; Prm, OPrm: string;
   Auth_Username, Auth_Password: string; FailedAuthAttempts: integer;
   HELOSent, SpoolAllocated, ReadSucceeded, UnexpectedFail: boolean;
   VStr: string; LogAgent: string;
   TempStr: string;

   procedure SendAndLogResponse(NumericCode: word; ReplyText: shortstring; ExpectFail: boolean = false);
   begin
      if (Logger.AddLine(LogAgent, 'Response: ' + IntToStr(NumericCode) + ' ' + ReplyText)) or ExpectFail then begin
         Response.SetReply(NumericCode, ReplyText);
         TCP.SendResponse(Response);
      end
      else begin
         SendAndLogResponse(SMTP_R_SERVICE_NA, 'Internal error: could not write log', true);
         Logger.AddStdLine(LogAgent, 'Log write failure. Terminating active connection.');
         UnexpectedFail:= true;
      end;
   end;

begin
   TCP:= Connection as TTCPRFCConnection;
   TCP.SetSockTimeOut(DEF_SOCK_TIMEOUT);
   TCP.ReverseDNSLookup;
   Originator:= TCP.HostIP.Copy;
   Response:= TRFCReply.Create;
   {PolicyObject:= PolicyManager.MakePolicyObject(Originator.Copy);}
   PolicyObject:= PolicyManager.MakePolicyObject(Originator);
   SpoolObject:= nil;
   HELOSent:= false; SpoolAllocated:= false; UnexpectedFail:= false;
   FailedAuthAttempts:= 0;

   { Prepare for logging. To make this connection distinguishable, we add
     the actual thread's ID to each log entry. }
   LogAgent:= 'Server ' + IntToStr(GetCurrentThreadId);
   Logger.AddLine(LogAgent, 'Client connected: ' + Originator.Name + ' (' + Originator.IP + ')');
   Logger.AddLine(LogAgent, 'Assigned rights (for host): ' + PolicyObject.RightsStr);

   { Verify FCrDNS if necessary. Note, maybe it would have been simpler to
     check it around the TCP.ReverseDNSLookup call, and then only pass the
     trusted result to PolicyManager.MakePolicyObject. The main idea why I
     didn't implement it that way is that I'd like to see if the granted
     rights actually change after the FCrDNS check. }
   if PolicyManager.FCrDNSPolicy <> FCRDNS_NAIVE then begin
      if not TCP.VerifyFCrDNS then begin
         PolicyManager.RevalidatePolicyObject(PolicyObject, Originator, false, PolicyManager.FCrDNSPolicy = FCRDNS_MEAN);
         if PolicyManager.FCrDNSPolicy = FCRDNS_STRICT then
            PolicyObject.Deny(RIGHT_CONNECT);
         Logger.AddLine(LogAgent, 'WARNING: "' + Originator.Name + '" is not a forward-confirmed reverse hostname! Rights will be reassigned by IP only!');
         Logger.AddLine(LogAgent, 'Assigned rights (for host): ' + PolicyObject.RightsStr);
      end;
   end;

   if PolicyObject.HasRight(RIGHT_CONNECT) then begin
      if not PolicyManager.HideVersion then VStr:= ' ' + MainServerConfig.VersionStr else VStr:= '';
      Response.SetReply(SMTP_R_READY, MainServerConfig.Name + ' SMTP server ready (MgSMTP' + VStr + ')');
      TCP.SendResponse(Response);

      repeat
         ReadSucceeded:= TCP.ReadCommand(Cmd, Prm);

         { Check if command only contains printable ASCII characters, not some binary garbage. }
         if ReadSucceeded then begin
            if IsPrintableString(Cmd) and IsPrintableString(Prm) then begin
               Logger.AddLine(LogAgent, 'Command: ' + Cmd + ' ' + Prm);
               Cmd:= UpperCase(Cmd);
            end
            else begin
               SendAndLogResponse(SMTP_R_SERVICE_NA, 'Non-printable characters are not allowed in SMTP commands! Stop abusing my service!');
               UnexpectedFail:= true;
            end;
         end;

         if (Length(Cmd) = 0) or (not ReadSucceeded) or UnexpectedFail then { Nothing. }

         else if (Cmd = 'GET') or (Cmd = 'HEAD') or (Cmd = 'POST') then begin
            SendAndLogResponse(SMTP_R_SERVICE_NA, 'Please learn to speak SMTP for I won''t speak HTTP. Stop abusing my service!');
            UnexpectedFail:= true;
         end

         else if (Cmd = SMTP_C_HELO) or (Cmd = SMTP_C_EHLO) then begin
            Response.SetReply(SMTP_R_OK, MainServerConfig.Name);
            if Cmd = SMTP_C_EHLO then begin
               Response.Add('SIZE ' + IntToStr(PolicyObject.Databytes));
               {Response.Add('VRFY');}
               Response.Add('PIPELINING');
               Response.Add('8BITMIME');
               if PolicyManager.Users then begin
                  Response.Add('AUTH LOGIN');
                  Response.Add('AUTH=LOGIN');
               end;
            end;
            TCP.SendResponse(Response);
            Originator.Free;
            Originator:= TIPNamePair.Create(Prm, TCP.HostIP.IP);
            HELOSent:= true;
            Logger.AddLine(LogAgent, 'Client identified: ' + Originator.Name + ' (' + Originator.IP + ')');
         end

         else if Cmd = SMTP_C_AUTH then begin
            if PolicyManager.Users then begin
               { Only "AUTH LOGIN" is supported. }
               SplitParameters(Prm, Prm, OPrm);
               if Prm = 'LOGIN' then begin
                  if OPrm = '' then begin
                     { Base64-encoded "Username:" }
                     Response.SetReply(SMTP_R_AUTH_MESSAGE, 'VXNlcm5hbWU6');
                     TCP.SendResponse(Response);
                     TCP.ReadLn(Auth_Username);
                     Auth_Username:= Base64Decode(Auth_Username);
                  end
                  else
                     Auth_Username:= Base64Decode(OPrm);
                  { Base64-encoded "Password:" }
                  Response.SetReply(SMTP_R_AUTH_MESSAGE, 'UGFzc3dvcmQ6');
                  TCP.SendResponse(Response);
                  TCP.ReadLn(Auth_Password);
                  { Verify }
                  if PolicyManager.AuthenticateUser(Auth_Username, Base64Decode(Auth_Password), PolicyObject) then begin
                     Response.SetReply(SMTP_R_AUTH_SUCCESSFUL, 'Authentication successful');
                     Logger.AddLine(LogAgent, 'Successfully authenticated as user: ' + Auth_Username);
                     Logger.AddLine(LogAgent, 'Assigned rights (for user): ' + PolicyObject.RightsStr);
                  end
                  else begin
                     Inc(FailedAuthAttempts);
                     Response.SetReply(SMTP_R_AUTH_FAILED, 'Authentication failed');
                     Logger.AddLine(LogAgent, 'AUTHENTICATION FAILED as user: ' + Auth_Username);
                  end;
                  TCP.SendResponse(Response);
                  if (PolicyManager.MaxAuthAttempts <> 0) and (PolicyManager.MaxAuthAttempts <= FailedAuthAttempts) then begin
                     SendAndLogResponse(SMTP_R_SERVICE_NA, 'Too many unsuccessful authentication attempts! Stop abusing my service!');
                     UnexpectedFail:= true;
                     Logger.AddLine(LogAgent, 'MAXIMUM AUTHENTICATION ATTEMPTS REACHED - DISCONNECTING CLIENT!');
                  end;
               end
               else
                  SendAndLogResponse(SMTP_R_PRM_NOT_IMPLEMENTED, 'Authentication type not implemented');
            end
            else
               SendAndLogResponse(SMTP_R_CMD_NOT_IMPLEMENTED, 'User authentication is not enabled on this server.');
         end

         else if Cmd = SMTP_C_RSET then begin
            { We must be careful to always free the spool object, if we
              have allocated one, but we don't need it anymore. }
            if SpoolAllocated then begin
               if SpoolObject.Opened then SpoolObject.Discard;
               SpoolObject.Free;
               SpoolAllocated:= false;
            end;
            Response.SetReply(SMTP_R_OK, 'OK');
            TCP.SendResponse(Response);
         end

         else if Cmd = SMTP_C_NOOP then begin
            Response.SetReply(SMTP_R_OK, 'Not like I was doing anything...');
            TCP.SendResponse(Response);
         end

         else if Cmd = SMTP_C_QUIT then begin
            { No extra action is required here to close the connection.
              The repeat-until loop will quit anyway, and the connection
              will be closed afterwards. }
            Response.SetReply(SMTP_R_CLOSE, 'Goodbye. :)');
            TCP.SendResponse(Response);
         end

         else if (HELOSent) or (not PolicyManager.ReqHELO) then begin

            { Some commands are only accepted after the client has greeted
              us with a HELO or EHLO command. }

            if Cmd = SMTP_C_MAIL then begin
               { A new spool object is allocated with the mail command. }
               if not SpoolAllocated then begin
                  OPrm:= Prm;
                  Prm:= CleanEMailAddress(Prm);
                  if (Prm = '') or (IsValidEMailAddress(Prm)) then begin
                     SpoolObject:= SpoolManager.CreateSpoolObject(Originator.Copy);
                     SpoolObject.Envelope.ReturnPath:= Prm;
                     SpoolObject.Databytes:= PolicyObject.Databytes;
                     SetEMailProperties(OPrm, SpoolObject);
                     if (SpoolObject.EMailProperties.Size <= SpoolObject.Databytes) then begin
                        Response.SetReply(SMTP_R_OK, 'OK');
                        TCP.SendResponse(Response);
                        SpoolAllocated:= true;
                        Logger.AddLine(LogAgent, 'Return-Path accepted: <' + Prm + '>');
                     end
                     else begin
                        SendAndLogResponse(SMTP_R_STOR_EXCEEDED, 'Declared message size exceeds the configured databytes limit');
                        SpoolObject.Free;
                     end;
                  end
                  else
                     SendAndLogResponse(SMTP_R_MB_SYNTAX_ERROR, '<' + Prm + '>: Sender address rejected: Syntax error');
               end
               else
                  SendAndLogResponse(SMTP_R_BAD_SEQUENCE, 'Return-Path is already specified, use RSET to discard it');
            end

            else if Cmd = SMTP_C_RCPT then begin
               if SpoolAllocated then begin
                  Prm:= CleanEMailAddress(Prm);

                  { According to the RFC, we must accept "POSTMASTER" address without a hostname. }
                  if UpperCase(Prm) = 'POSTMASTER' then Prm:= Prm + '@' + MainServerConfig.Name;
                  if IsValidEMailAddress(Prm) then begin

                     if MailboxManager.IsLocalAddress(Prm) then begin

                        { Many conditions need to be checked before accepting a local e-mail:
                          - Does this server accept local e-mails by configuration?
                          - Does the client have the right to STORE a local e-mail?
                          - Does the addressed mailbox exist?
                          - Does the mailbox have free quota?
                          If the answer is "no" for any of these questions, reject the address
                          with a proper error response. }

                        if MainServerConfig.Mailbox then begin
                           if PolicyObject.HasRight(RIGHT_STORE) then begin
                              if MailboxManager.Verify(Prm) then begin
                                 if MailboxManager.VerifyAlias(Prm) then begin
                                    if ((not SpoolManager.AllowExceedQuota) and (MailboxManager.CheckQuota(EMailUserName(Prm), EMailHost(Prm), SpoolObject.EMailProperties.Size)))
                                    or ((SpoolManager.AllowExceedQuota) and (MailboxManager.CheckQuota(EMailUserName(Prm), EMailHost(Prm), 0))) then begin

                                       if MailboxManager.Rewrite then begin
                                          TempStr:= HandleRewrite(Prm, MailboxManager.GetMailbox(EMailUserName(Prm), EMailHost(Prm)), SpoolObject);
                                          if Length(TempStr) > 0 then
                                             Logger.AddLine(LogAgent, TempStr);
                                       end
                                       else
                                          SpoolObject.Envelope.AddRecipient(Prm);

                                       Response.SetReply(SMTP_R_OK, 'OK');
                                       TCP.SendResponse(Response);
                                       Logger.AddLine(LogAgent, 'Local recipient accepted: <' + Prm + '>');
                                    end
                                    else
                                       SendAndLogResponse(SMTP_R_STOR_EXCEEDED, '<' + Prm + '>: User quota exceeded');
                                 end
                                 else
                                    SendAndLogResponse(SMTP_R_MAILBOX_NA, '<' + Prm + '>: Mailbox alias rejected');
                              end
                              else
                                 SendAndLogResponse(SMTP_R_MAILBOX_NA, '<' + Prm + '>: No mailbox here by that name');
                           end
                           else
                              SendAndLogResponse(SMTP_R_MAILBOX_NA, '<' + Prm + '>: Store access denied');
                        end
                        else
                           SendAndLogResponse(SMTP_R_MAILBOX_NA, '<' + Prm + '>: This server doesn''t store local messages');
                     end

                     else if MainServerConfig.Relay then begin

                        { Things to check for relay addresses:
                          - Does the server ever accept relay addresses by configuration?
                          - Does the client has the right to RELAY messages or in the case
                            if the relay address is on the RelayTo list, does the client
                            has the STORE right?
                        }

                        if (PolicyObject.HasRight(RIGHT_RELAY))
                           or (PolicyObject.HasRight(RIGHT_STORE) and RelayManager.IsOnRelayToList(EMailHost(Prm))) then begin
                           if not RelayManager.IsOnNoRelayToList(EMailHost(Prm)) then begin
                              SpoolObject.Envelope.AddRecipient(Prm);
                              Response.SetReply(SMTP_R_OK, 'OK');
                              TCP.SendResponse(Response);
                              Logger.AddLine(LogAgent, 'Relay recipient accepted: <' + Prm + '>');
                           end
                           else
                              SendAndLogResponse(SMTP_R_TRANS_FAILED, '<' + Prm + '>: Relaying towards this domain is not permitted');
                        end
                        else
                           SendAndLogResponse(SMTP_R_TRANS_FAILED, '<' + Prm + '>: Relay access denied, or maybe I just don''t like you');
                     end
                     else
                        SendAndLogResponse(SMTP_R_TRANS_FAILED, '<' + Prm + '>: Relaying has been disabled by configuration');
                  end
                  else
                     SendAndLogResponse(SMTP_R_MB_SYNTAX_ERROR, '<' + Prm + '>: Recipient address rejected: Syntax error');
               end
               else
                  SendAndLogResponse(SMTP_R_BAD_SEQUENCE, 'You must initiate e-mail transactions with MAIL command');
            end

            else if Cmd = SMTP_C_DATA then begin
               if SpoolAllocated then begin
                  if SpoolObject.Envelope.IsComplete then begin
                     ReceiveEMailData(TCP, Response, SpoolObject);
                     Logger.AddLine(LogAgent, 'Response: ' + IntToStr(Response.NumericCode) + ' ' + Response.GetLine(0));
                     TCP.SendResponse(Response);
                     Logger.AddLine('Object ' + SpoolObject.Name, 'Message-ID: <' + SpoolObject.OriginalMessageID + '>');
                     SpoolObject.Free;
                     SpoolAllocated:= false;
                  end
                  else
                     SendAndLogResponse(SMTP_R_TRANS_FAILED, 'No valid recipients');
               end
               else
                  SendAndLogResponse(SMTP_R_BAD_SEQUENCE, 'You must initiate e-mail transactions with MAIL command');
            end

            else if Cmd = SMTP_C_VRFY then
               SendAndLogResponse(SMTP_R_CANNOTVERIFY, 'Honestly, I don''t like to verify addresses')

            else
               SendAndLogResponse(SMTP_R_CMD_SYNTAX_ERROR, 'Command not recognized (' + Cmd + ')');
         end

         else
            SendAndLogResponse(SMTP_R_BAD_SEQUENCE, 'It would be more polite to say HELO first');

      until (Cmd = SMTP_C_QUIT) or (not ReadSucceeded) or (UnexpectedFail);

      if not ReadSucceeded then
         SendAndLogResponse(SMTP_R_SERVICE_NA, 'Socket read error');
   end

   else begin

      { If the client doesn't have the right to CONNECT here, disconnect it
        with a rather unfriendly message. }

      SendAndLogResponse(SMTP_R_TRANS_FAILED, 'Host is not permitted by server configuration');
      SendAndLogResponse(SMTP_R_SERVICE_NA, 'You are not welcome here, I shall disconnect you');
      {repeat
         TCP.ReadCommand(Cmd, Prm);
         if Cmd <> SMTP_C_QUIT then
            Response.SetReply(SMTP_R_BAD_SEQUENCE, 'You are not welcome here, I suggest you to QUIT')
         else
            Response.SetReply(SMTP_R_CLOSE, 'Closing connection');
         TCP.SendResponse(Response);
      until Cmd = SMTP_C_QUIT;}
   end;

   { Free the spool object (if we have any), close the connection,
     and free other allocated resources, log disconnection. }

   if SpoolAllocated then begin
      if SpoolObject.Opened then SpoolObject.Discard;
      SpoolObject.Free;
   end;
   PolicyObject.Free;
   Response.Free;
   Originator.Free;
   TCP.Free;
   Logger.AddLine(LogAgent, 'Client disconnected.');
end;

procedure TMgSMTPListener.ReceiveEMailData(TCP: TTCPRFCConnection; Response: TRFCReply; SpoolObject: TSpoolObjectCreator);
{ Receive e-mail lines until a line with a single dot (".") arrives.
  Check databytes limit!
  This procedure should never call TCP.SendResponse - the set up response
  will be sent by the caller! }
var Line: string; Done, ReadOK: boolean;
begin
   if SpoolObject.Open then begin
      Response.SetReply(SMTP_R_START_MAIL_INPUT, 'Start mail input; end with "<CRLF>.<CRLF>" sequence');
      TCP.SendResponse(Response);
      Done:= false;
      repeat
         ReadOK:= TCP.ReadLn(Line);
         if Line <> '.' then begin
            { If the line starts with a dot, remove it to comply with RFC. }
            if (Length(Line) > 1) and (Line[1] = '.') then Delete(Line, 1, 1);
            SpoolObject.DeliverMessagePart(Line);
         end
         else
            Done:= true;
      until Done or (not ReadOK);
      if ReadOK then begin
         if SpoolObject.GetErrorCode <> SCE_NO_ERROR then begin

            case SpoolObject.GetErrorCode of

               SCE_SIZE_EXCEEDED:
                  Response.SetReply(SMTP_R_STOR_EXCEEDED, 'Message size exceeds the configured databytes limit');

               SCE_LOOP_DETECTED:
               begin
                  Response.Clear;
                  Response.SetNumericCode(SMTP_R_TRANS_FAILED);
                  Response.Add('Too many "Received" headers in mail data.');
                  Response.Add('It''s likely that your message got trapped in a mail relay loop. In most');
                  Response.Add('cases it is caused by faulty mail server configuration. Please notify the');
                  Response.Add('administrator by forwarding this failure notice to the following address:');
                  Response.Add('<postmaster@' + MainServerConfig.Name + '>!');
               end;

               SCE_WRITE_FAIL:
                  Response.SetReply(SMTP_R_ABORTED, 'Could not write mail data. Try again later.');

               else
                  Response.SetReply(SMTP_R_ABORTED, 'Unknown error. Could not queue mail data.');

            end;

            SpoolObject.Discard;

         end
         else begin
            Response.SetReply(SMTP_R_OK, 'Queued as ' + SpoolObject.Name);
            SpoolObject.Close;
         end;
      end
      else begin
         Response.SetReply(SMTP_R_SERVICE_NA, 'Socket read error in DATA phase (timeout?)');
         SpoolObject.Discard;
      end;
   end
   else Response.SetReply(SMTP_R_ABORTED, 'Internal error: could not open spool object');
end;


end.
