{
   MegaBrutal's SMTP Server (MgSMTP)
   Copyright (C) 2010-2015 MegaBrutal

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
}

{
   Unit: Mailbox
   Administers mailboxes. It implements methods to place messages into the
   mailbox files.
}


{$MODE DELPHI}

unit Mailbox;

interface
uses SysUtils, SyncObjs, Classes, INIFiles, Common, Log, Spool;

type

   PMailbox = ^TMailbox;
   TMailbox = class(TNamedObject)
      constructor Create(const Name, Domain: string; Config: TINIFile; Slave: boolean; DefaultQuota: longint);
      destructor Destroy; override;
   protected
      FDomain: string;
      FQuota, FLockID: longint;
      CriticalSection: TCriticalSection;
   private
      FPlusAliases: boolean;
      FRewritePassThru: boolean;
      PlusAliasExceptList: TStringList;
      RewriteToList: TStringList;
      MailboxFile: TStream;
      procedure AddTrackHeaders(EMail, Recipient: string; Headers: TStrings);
      { class function GetConfigSectionName(const Name, Domain: string): string; }
      class function GetMailboxConfig(Config: TINIFile; const Name, Domain, Ident, Default: string): string; overload;
      class function GetMailboxConfig(Config: TINIFile; const Name, Domain, Ident: string; Default: boolean): boolean; overload;
   public
      function IsItYourName(const Name: string): boolean; override;
      function GetMailboxAddress: string;
      function CheckAlias(const Name: string): boolean;
      function CheckQuota(MailSize: longint): boolean; virtual; abstract;
      function BeginDeliverMessage(LockID: longint; EMail, Recipient, SpoolID: string; EMailProperties: TEMailProperties; Headers: TStrings): boolean; virtual; abstract;
      function DeliverMessagePart(LockID: longint; Message: TStrings): boolean; virtual; abstract;
      function FinishDeliverMessage(LockID: longint): boolean; virtual; abstract;
      function Lock: longint; virtual; abstract;
      function Release(LockID: longint): boolean; virtual; abstract;
      function GetRewriteCount: integer;
      function GetRewriteToEntry(i: integer): string;
      function GetRewriteToListStr: string;
      property Domain: string read FDomain;
      property Quota: longint read FQuota;
      property RewriteCount: integer read GetRewriteCount;
      property RewritePassThru: boolean read FRewritePassThru;
   end;

   TMailbox_mbox = class(TMailbox)
   private
      procedure FromQuote(var Message: TStrings);
      function MakeMailboxFilename: string;
   public
      function CheckQuota(MailSize: longint): boolean; override;
      function BeginDeliverMessage(LockID: longint; EMail, Recipient, SpoolID: string; EMailProperties: TEMailProperties; Headers: TStrings): boolean; override;
      function DeliverMessagePart(LockID: longint; Message: TStrings): boolean; override;
      function FinishDeliverMessage(LockID: longint): boolean; override;
      function Lock: longint; override;
      function Release(LockID: longint): boolean; override;
   end;

   TForwarderMailbox = class(TMailbox)
      constructor Create(const Name, Domain: string; Config: TINIFile; Slave: boolean; DefaultQuota: longint; PhysicalMailbox: TMailbox);
      destructor Destroy; override;
   private
      PhysicalMailbox: TMailbox;
      ForwardToList: TStringList;
      SpoolObject: TSpoolObjectCreator;
      OrigSpoolID: string;
      FReplyTo: string;
      FForwardHeaders, FRemail: boolean;
   public
      function CheckQuota(MailSize: longint): boolean; override;
      function BeginDeliverMessage(LockID: longint; EMail, Recipient, SpoolID: string; EMailProperties: TEMailProperties; Headers: TStrings): boolean; override;
      function DeliverMessagePart(LockID: longint; Message: TStrings): boolean; override;
      function FinishDeliverMessage(LockID: longint): boolean; override;
      function Lock: longint; override;
      function Release(LockID: longint): boolean; override;
      property ReplyTo: string read FReplyTo;
      property ForwardHeaders: boolean read FForwardHeaders;
      property Remail: boolean read FRemail;
   end;


   TBoxes = array of TMailbox;
   TDomainBoxes = array of TBoxes;

   TMailboxContainer = class(TStringList)
   public
      destructor Destroy; override;
   protected
      DomainBoxes: TDomainBoxes;
   public
      procedure AddMailbox(const Domain: string; Mailbox: TMailbox);
      function GetMailbox(const Name, Domain: string): PMailbox;
   end;


   TMailboxManager = class
      constructor Create(Config: TINIFile);
      destructor Destroy; override;
   private
      MailboxContainer: TMailboxContainer;
      DefaultQuota: longint;
      FDomainSpecific: boolean;
      FRewrite, FForward: boolean;
   public
      property DomainSpecific: boolean read FDomainSpecific;
      property Rewrite: boolean read FRewrite;
      property Forward: boolean read FForward;
      function CheckQuota(const Name, Domain: string; MailSize: longint): boolean;
      function GetMailbox(const Name, Domain: string): PMailbox;
      function IsLocalAddress(const EMail: string): boolean;
      function Verify(const EMail: string): boolean;
      function VerifyAlias(const EMail: string): boolean;
   end;


var

   MailboxManager: TMailboxManager;


implementation

const

   { Search attributes: }
   SEARCH_ATTR = faAnyFile - faDirectory - faVolumeID - faHidden;


constructor TMailbox.Create(const Name, Domain: string; Config: TINIFile; Slave: boolean; DefaultQuota: longint);
var Section, BaseList: string;
begin
   if Length(Domain) = 0 then
      Section:= 'Mailbox\' + Name
   else
      Section:= 'Mailbox\' + Name + '@' + Domain;

   inherited Create(Name, Config, Section);
   FDomain:= Domain;
   FQuota:= Config.ReadInteger(Section, 'Quota', DefaultQuota);
   FQuota:= StrToIntDef(GetMailboxConfig(Config, Name, Domain, 'Quota', IntToStr(DefaultQuota)), DefaultQuota);
   FLockID:= 0;
   CriticalSection:= TCriticalSection.Create;

   if (not Slave) then begin
      FPlusAliases:= GetMailboxConfig(Config, Name, Domain, 'PlusAliases', true);
      if FPlusAliases then begin
         PlusAliasExceptList:= TStringList.Create;
         PlusAliasExceptList.Delimiter:= ',';
         BaseList:= GetMailboxConfig(Config, Name, Domain, 'PlusAliasExcept', '');
         if Length(Domain) > 0 then
            PlusAliasExceptList.DelimitedText:= BaseList + ',' + Config.ReadString('Mailbox\@' + Domain, 'GlobalPlusAliasExcept', '')
         else
            PlusAliasExceptList.DelimitedText:= BaseList + ',' + Config.ReadString('Mailbox', 'GlobalPlusAliasExcept', '');

     end;
   end;

   if (not Slave) and Config.ReadBool('Mailbox', 'Rewrite', false) then begin
      RewriteToList:= TStringList.Create;
      RewriteToList.Delimiter:= ',';
      RewriteToList.DelimitedText:= GetMailboxConfig(Config, Name, Domain, 'RewriteTo', '');
      FRewritePassThru:= GetMailboxConfig(Config, Name, Domain, 'RewritePassThru', true);
   end
   else FRewritePassThru:= true;
end;

destructor TMailbox.Destroy;
begin
   CriticalSection.Free;
end;

constructor TForwarderMailbox.Create(const Name, Domain: string; Config: TINIFile; Slave: boolean; DefaultQuota: longint; PhysicalMailbox: TMailbox);
begin
   inherited Create(Name, Domain, Config, Slave, DefaultQuota);
   Self.PhysicalMailbox:= PhysicalMailbox;

   FReplyTo:= GetMailboxConfig(Config, Name, Domain, 'ReplyTo', '');
   FForwardHeaders:= GetMailboxConfig(Config, Name, Domain, 'ForwardHeaders', true);
   FRemail:= GetMailboxConfig(Config, Name, Domain, 'Remail', false);

   ForwardToList:= TStringList.Create;
   ForwardToList.Delimiter:= ',';
   ForwardToList.DelimitedText:= GetMailboxConfig(Config, Name, Domain, 'ForwardTo', '');
end;

destructor TForwarderMailbox.Destroy;
begin
   inherited Destroy;
   ForwardToList.Free;
   if (PhysicalMailbox <> nil) then PhysicalMailbox.Free;
end;

destructor TMailboxContainer.Destroy;
var i, j: integer;
begin
   for i:= Length(DomainBoxes) - 1 downto 0 do begin
      for j:= Length(DomainBoxes[i]) - 1 downto 0 do begin
         DomainBoxes[i][j].Free;
      end;
   end;
   inherited Destroy;
end;

constructor TMailboxManager.Create(Config: TINIFile);
var SearchRec: TSearchRec; i: integer; BoxName, BoxDomain: string;
    SlaveMailbox: TMailBox;
begin
   inherited Create;
   DefaultQuota:=       Config.ReadInteger('Mailbox', 'Quota', 0);
   FDomainSpecific:=    Config.ReadBool('Mailbox', 'DomainSpecific', false);
   FRewrite:=           Config.ReadBool('Mailbox', 'Rewrite', false);
   FForward:=           Config.ReadBool('Mailbox', 'Forward', false);
   MailboxContainer:=   TMailboxContainer.Create;
   if FindFirst('mail\*', SEARCH_ATTR, SearchRec) = 0 then begin
      i:= 0;
      repeat
         if DomainSpecific then begin
            BoxName:= EMailUserName(SearchRec.Name);
            BoxDomain:= EMailHost(SearchRec.Name);
         end
         else begin
            BoxName:= SearchRec.Name;
            BoxDomain:= '';
         end;

         { If forwarding requested, set up a forwarder mailbox. }

         if Forward and (Length(TMailbox.GetMailboxConfig(Config, BoxName, BoxDomain, 'ForwardTo', '')) > 0) then begin

            if TMailbox.GetMailboxConfig(Config, BoxName, BoxDomain, 'StoreLocalCopy', true) then
               SlaveMailbox:= TMailbox_mbox.Create(BoxName, BoxDomain, Config, true, DefaultQuota)
            else
               SlaveMailbox:= nil;

            MailboxContainer.AddMailbox(BoxDomain, TForwarderMailbox.Create(BoxName, BoxDomain, Config, false, DefaultQuota, SlaveMailbox));
         end
         else
            MailboxContainer.AddMailbox(BoxDomain, TMailbox_mbox.Create(BoxName, BoxDomain, Config, false, DefaultQuota));

         Inc(i);
      until FindNext(SearchRec) <> 0;
   end;
   FindClose(SearchRec);
end;

destructor TMailboxManager.Destroy;
begin
   MailboxContainer.Free;
   inherited Destroy;
end;


procedure TMailbox.AddTrackHeaders(EMail, Recipient: string; Headers: TStrings);
begin
   Headers.Insert(0, 'Return-Path: <' + EMail + '>');
   Headers.Insert(0, 'X-Original-To: <' + Recipient + '>');
end;

{class function TMailbox.GetConfigSectionName(const Name, Domain: string): string;
begin
   if Length(Domain) = 0 then
      Result:= 'Mailbox\' + Name
   else
      Result:= 'Mailbox\' + Name + '@' + Domain;
end;}

class function TMailbox.GetMailboxConfig(Config: TINIFile; const Name, Domain, Ident, Default: string): string;
begin
   if Length(Domain) > 0 then
      Result:= Config.ReadString('Mailbox\' + Name + '@' + Domain, Ident,
         Config.ReadString('Mailbox\@' + Domain, Ident,
         Config.ReadString('Mailbox', Ident, Default)))
   else
      Result:= Config.ReadString('Mailbox\' + Name, Ident,
         Config.ReadString('Mailbox', Ident, Default));
end;

class function TMailbox.GetMailboxConfig(Config: TINIFile; const Name, Domain, Ident: string; Default: boolean): boolean;
begin
   if Length(Domain) > 0 then
      Result:= Config.ReadBool('Mailbox\' + Name + '@' + Domain, Ident,
         Config.ReadBool('Mailbox\@' + Domain, Ident,
         Config.ReadBool('Mailbox', Ident, Default)))
   else
      Result:= Config.ReadBool('Mailbox\' + Name, Ident,
         Config.ReadBool('Mailbox', Ident, Default));
end;

function TMailbox.IsItYourName(const Name: string): boolean;
var p: integer;
begin
   if FPlusAliases then begin
      p:= Pos('+', Name);
      if p <> 0 then
         Result:= inherited IsItYourName(Copy(Name, 1, p - 1))
      else
         Result:= inherited IsItYourName(Name);
   end
   else
      Result:= inherited IsItYourName(Name);
end;

function TMailbox.GetMailboxAddress: string;
begin
   if Length(Domain) = 0 then
      Result:= Name + '@' + MainServerConfig.Name
   else
      Result:= Name + '@' + Domain;
end;

function TMailbox.CheckAlias(const Name: string): boolean;
var p: integer;
begin
   p:= Pos('+', Name);
   if p <> 0 then
      Result:= PlusAliasExceptList.IndexOf(Copy(Name, p + 1, Length(Name) - p)) = -1
   else
      Result:= true;
end;

function TMailbox.GetRewriteCount: integer;
begin
   Result:= RewriteToList.Count;
end;

function TMailbox.GetRewriteToEntry(i: integer): string;
begin
   Result:= RewriteToList.Strings[i];
end;

function TMailbox.GetRewriteToListStr: string;
begin
   Result:= RewriteToList.DelimitedText;
end;


procedure TMailbox_mbox.FromQuote(var Message: TStrings);
var i: integer;
begin
   for i:= 0 to Message.Count - 1 do
      if pos('From ', Message.Strings[i]) = 1 then
         Message.Strings[i]:= '>' + Message.Strings[i];
end;

function TMailbox_mbox.MakeMailboxFilename: string;
begin
   if Length(Domain) = 0 then
      Result:= 'mail\' + Name
   else
      Result:= 'mail\' + Name + '@' + Domain;
end;

function TMailbox_mbox.CheckQuota(MailSize: longint): boolean;
{ Returns FALSE if the given message size would exceed the quota. }
var SearchRec: TSearchRec;
begin
   if FindFirst(MakeMailboxFilename, SEARCH_ATTR, SearchRec) = 0 then
      Result:= ((SearchRec.Size + MailSize) <= FQuota) or (FQuota = 0)
   else
      Result:= false;
   FindClose(SearchRec);
end;

function TMailbox_mbox.BeginDeliverMessage(LockID: longint; EMail, Recipient, SpoolID: string; EMailProperties: TEMailProperties; Headers: TStrings): boolean;
var NL, Line: string;
begin
   if (FLockID <> 0) and (FLockID = LockID) then begin
      case DefaultTextLineBreakStyle of
         tlbsLF:     NL:= #10;
         tlbsCRLF:   NL:= #13#10;
         tlbsCR:     NL:= #13;
      end;

      {FromQuote(Headers);}
      AddTrackHeaders(EMail, Recipient, Headers);
      Headers.Insert(0, 'Delivered-To: ' + GetMailboxAddress);
      try
         if Length(EMail) = 0 then EMail:= 'MAILER-DAEMON';
         Line:= 'From ' + EMail + ' ' + EMailTimeStamp(Now) + NL;
         MailboxFile.WriteBuffer(Pointer(Line)^,Length(Line));
         Headers.SaveToStream(MailboxFile);
         MailboxFile.WriteBuffer(Pointer(NL)^,Length(NL));
         Result:= true;
      except
         Result:= false;
      end;
   end
   else Result:= false;
end;

function TMailbox_mbox.DeliverMessagePart(LockID: longint; Message: TStrings): boolean;
begin
   if (FLockID <> 0) and (FLockID = LockID) then begin
      FromQuote(Message);
      try
         Message.SaveToStream(MailboxFile);
         Result:= true;
      except
         Result:= false;
      end;
   end
   else Result:= false;
end;

function TMailbox_mbox.FinishDeliverMessage(LockID: longint): boolean;
begin
   if (FLockID <> 0) and (FLockID = LockID) then begin
      try
         MailboxFile.WriteBuffer(#13#10, 2);
         Result:= true;
      except
         Result:= false;
      end;
   end
   else Result:= false;
end;

function TMailbox_mbox.Lock: longint;
begin
   CriticalSection.Acquire;
   if FLockID = 0 then begin
      try
         MailboxFile:= TFileStream.Create(MakeMailboxFilename, fmOpenReadWrite);
         FLockID:= (MailboxFile as TFileStream).Handle;
         MailboxFile.Seek(0, soFromEnd);
      except
         try
            FreeAndNil(MailboxFile);
         except
         end;
         FLockID:= 0;
      end;
      Result:= FLockID;
   end
   else Result:= 0;
   CriticalSection.Release;
end;

function TMailbox_mbox.Release(LockID: longint): boolean;
begin
   CriticalSection.Acquire;
   if (FLockID <> 0) and (FLockID = LockID) then begin
      FreeAndNil(MailboxFile);
      FLockID:= 0;
   end
   else Result:= false;
   CriticalSection.Release;
end;


function TForwarderMailbox.CheckQuota(MailSize: longint): boolean;
{ Returns FALSE if the given message size would exceed the quota. }
begin
   if PhysicalMailbox <> nil then
      Result:= PhysicalMailbox.CheckQuota(MailSize)
   else
      Result:= true;
end;

function TForwarderMailbox.BeginDeliverMessage(LockID: longint; EMail, Recipient, SpoolID: string; EMailProperties: TEMailProperties; Headers: TStrings): boolean;
var i: integer;
begin
   if (FLockID <> 0) and (FLockID = LockID) then begin
      if PhysicalMailbox = nil then begin
         AddTrackHeaders(EMail, Recipient, Headers);
         Result:= true;
      end
      else begin
         Result:= PhysicalMailbox.BeginDeliverMessage(LockID, EMail, Recipient, SpoolID, EMailProperties, Headers);
      end;

      if Result then begin
         SpoolObject:= SpoolManager.CreateSpoolObject(TIPNamePair.Create('internal', ''));
         //PrepareSpoolObject(EMail, ForwardSpoolObject, Headers);
         SpoolObject.Databytes:= 0;

         SpoolObject.EMailProperties.Size:=  EMailProperties.Size;
         SpoolObject.EMailProperties.Flags:= EMailProperties.Flags;
         OrigSpoolID:= SpoolID;

         { Forward or remail? Regardless of the settings, DSNs only get forwarded
           and never get remailed. (Remailing them could cause a loop.) }
         if (not Remail) or (EMail = '') then
            SpoolObject.Envelope.ReturnPath:= EMail
         else
            SpoolObject.Envelope.ReturnPath:= Recipient;

         for i:= 0 to ForwardToList.Count - 1 do
            SpoolObject.Envelope.AddRecipient(ForwardToList.Strings[i]);

         if ForwardHeaders then begin
            Headers.Insert(0, 'X-Forwarded-For: ' + Recipient + ' ' + ForwardToList.DelimitedText);
            Headers.Insert(0, 'X-Forwarded-To: ' + ForwardToList.DelimitedText);
         end;

         if ReplyTo <> '' then begin
            if ReplyTo = '!' then
               Headers.Insert(0, 'Reply-To: <' + Recipient + '>')
            else
               Headers.Insert(0, 'Reply-To: ' + ReplyTo);
         end;

         SpoolObject.Open;
         for i:= 0 to Headers.Count - 1 do
            SpoolObject.DeliverMessagePart(Headers.Strings[i]);

         SpoolObject.DeliverMessagePart('');
      end;
   end
   else
      Result:= false;
end;

function TForwarderMailbox.DeliverMessagePart(LockID: longint; Message: TStrings): boolean;
var i: integer;
begin
   if (FLockID <> 0) and (FLockID = LockID) then begin
      if PhysicalMailbox <> nil then
         Result:= PhysicalMailbox.DeliverMessagePart(LockID, Message)
      else
         Result:= true;

      for i:= 0 to Message.Count - 1 do
         SpoolObject.DeliverMessagePart(Message.Strings[i]);
   end
   else
      Result:= false;
end;

function TForwarderMailbox.FinishDeliverMessage(LockID: longint): boolean;
var action: string;
begin
   if (FLockID <> 0) and (FLockID = LockID) then begin
      if PhysicalMailbox <> nil then
         Result:= PhysicalMailbox.FinishDeliverMessage(LockID)
      else
         Result:= true;

      if not Remail then action:= 'forwarding' else action:= 'remailing';

      if SpoolObject.GetErrorCode = SCE_NO_ERROR then begin
         Logger.AddLine('Mailbox ' + GetMailboxAddress, 'Message ' + OrigSpoolID + ' has been copied to '
            + SpoolObject.Name + ' for ' + action + ' to ' + ForwardToList.DelimitedText);
         SpoolObject.Close;
      end
      else begin
         Logger.AddLine('Mailbox ' + GetMailboxAddress, 'Failed to copy message <' + SpoolObject.OriginalMessageID + '> for '
            + action + ' to ' + ForwardToList.DelimitedText
            + '; Spool error code: ' + IntToStr(SpoolObject.GetErrorCode));
         SpoolObject.Discard;
      end;
      SpoolObject.Free;
   end
   else
      Result:= false;
end;

function TForwarderMailbox.Lock: longint;
{ Very-very sensitive method that induced lots of cursing.
  Pay a lot of attention when you do any change to it! }
begin
   CriticalSection.Acquire;
   if FLockID = 0 then begin
      if PhysicalMailbox <> nil then begin
         FLockID:= PhysicalMailbox.Lock;
         Result:= FLockID;
      end
      else begin
         FLockID:= GetCurrentThreadID;
         Result:= FLockID;
      end;
   end
   else Result:= 0;
   CriticalSection.Release;
end;

function TForwarderMailbox.Release(LockID: longint): boolean;
begin
   CriticalSection.Acquire;
   if (FLockID <> 0) and (FLockID = LockID) then begin
      if PhysicalMailbox <> nil then
         Result:= PhysicalMailbox.Release(LockID)
      else
         Result:= true;
   end
   else
      Result:= false;
   if Result then FLockID:= 0;
   CriticalSection.Release;
end;


procedure TMailboxContainer.AddMailbox(const Domain: string; Mailbox: TMailbox);
var i, j: integer;
begin
   { IndexOf is supposed to be case-insensitive, but it depends on locales.
     It's safer to uppercase the string before passing to it.
     See: http://62.166.198.202/view.php?id=15489 }
   i:= IndexOf(UpperCase(Domain));
   if i = -1 then begin
      Add(Domain);
      i:= Count - 1;
      SetLength(DomainBoxes, Count);
   end;
   j:= Length(DomainBoxes[i]);
   SetLength(DomainBoxes[i], j + 1);
   DomainBoxes[i][j]:= Mailbox;
end;

function TMailboxContainer.GetMailbox(const Name, Domain: string): PMailbox;
var i, j: integer;
begin
   i:= IndexOf(UpperCase(Domain));
   if i <> -1 then begin
      j:= 0;
      while (j < Length(DomainBoxes[i])) and (not(DomainBoxes[i][j].IsItYourName(Name))) do Inc(j);
      if (j < Length(DomainBoxes[i])) then Result:= @DomainBoxes[i][j]
      else if Domain <> '' then Result:= GetMailbox(Name, '')
      else Result:= nil;
   end
   else if Domain <> '' then Result:= GetMailbox(Name, '')
   else Result:= nil;
end;


function TMailboxManager.CheckQuota(const Name, Domain: string; MailSize: longint): boolean;
var Mailbox: PMailbox;
begin
   Mailbox:= GetMailbox(Name, Domain);
   if Mailbox <> nil then
      Result:= Mailbox^.CheckQuota(MailSize)
   else
      Result:= false;
end;

function TMailboxManager.GetMailbox(const Name, Domain: string): PMailbox;
begin
   Result:= MailboxContainer.GetMailbox(Name, Domain);
end;

function TMailboxManager.IsLocalAddress(const EMail: string): boolean;
begin
   Result:= MainServerConfig.IsItYourName(EMailHost(EMail));
end;

function TMailboxManager.Verify(const EMail: string): boolean;
begin
   Result:= GetMailbox(EMailUserName(EMail), EMailHost(EMail)) <> nil;
end;

function TMailboxManager.VerifyAlias(const EMail: string): boolean;
var Mailbox: PMailbox;
begin
   Mailbox:= GetMailbox(EMailUserName(EMail), EMailHost(EMail));
   if Mailbox <> nil then
      Result:= Mailbox^.CheckAlias(EMailUserName(EMail))
   else
      Result:= false;
end;


end.
