{
   MegaBrutal's SMTP Server (MgSMTP)
   Copyright (C) 2010-2015 MegaBrutal

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
}


{$MODE DELPHI}
{$APPTYPE CONSOLE}

program MgSMTP;
uses Windows, SysUtils, INIFiles, EINIFiles,
     Common, Log, Spool, Listener, Mailbox, Relay, Policies;


function RegisterServiceCtrlHandlerEx(lpServiceName: LPCSTR; lpHandlerProc: pointer; lpContext: LPVOID): SERVICE_STATUS_HANDLE; stdcall; external 'advapi32.dll' name 'RegisterServiceCtrlHandlerExA';
procedure Service(Argc: dword; Argv: pointer); stdcall; forward;

const

   ServiceTable: array[0..1] of TServiceTableEntry =
      (
         (
            lpServiceName: 'MgSMTP';
            lpServiceProc: @Service;
         ),
         (
            lpServiceName: nil;
            lpServiceProc: nil;
         )
      );

   ArgumentPrefixes: array[0..2] of string = ('/', '--', '-');
   ValidArguments: array[0..4] of string = ('?', 'HELP', 'INSTALL', 'UNINSTALL', 'USERMODE');

   { For development test builds, you can add a developer comment here to
     document what bugfix/feature are you testing with the actual build.
     This will be logged to help you differentiate outputs of subsequent
     builds in your logs. If left empty, it won't be added to the logs. }
   DEVCOMMENT  =  'Detect unexpected disconnection';

var

   Cmdline: TArgumentParser;
   Config: TINIFile;
   hSCManager, hService: THandle;
   hSvcStatusHandle: THandle;
   SvcStatus: TServiceStatus;
   ServiceMode, Stopping: boolean;
   WrongArgument: integer;


procedure AddDevComment(Log: TStreamLogger);
begin
   if DEVCOMMENT <> '' then begin
      Log.writeln('Build: ' + {$INCLUDE %DATE%} + ' ' + {$INCLUDE %TIME%});
      Log.writeln('Developer note: ' + DEVCOMMENT);
   end;
end;

procedure ReportSvcStatus(CurrentState, WinExitCode, WaitHint: dword);
{ Report a service status to the Windows SCM. }
begin
   if ServiceMode then begin
      SvcStatus.dwCurrentState:= CurrentState;
      SvcStatus.dwWaitHint:= WaitHint;
      if WinExitCode <> 0 then begin
         SvcStatus.dwWin32ExitCode:= ERROR_SERVICE_SPECIFIC_ERROR;
         SvcStatus.dwServiceSpecificExitCode:= WinExitCode;
      end
      else begin
         SvcStatus.dwWin32ExitCode:= NO_ERROR;
      end;

      if CurrentState = SERVICE_START_PENDING then
         SvcStatus.dwControlsAccepted:= 0
      else begin
         if GetWinMajorVersion >= 6 then begin
            SvcStatus.dwControlsAccepted:=
               SERVICE_ACCEPT_STOP or SERVICE_ACCEPT_SHUTDOWN or SERVICE_ACCEPT_PRESHUTDOWN;
         end
         else begin
            SvcStatus.dwControlsAccepted:=
               SERVICE_ACCEPT_STOP or SERVICE_ACCEPT_SHUTDOWN;
         end;
      end;

      if (CurrentState = SERVICE_RUNNING) or (CurrentState = SERVICE_STOPPED) then
         SvcStatus.dwCheckPoint:= 0
      else
         Inc(SvcStatus.dwCheckPoint);

      SetServiceStatus(hSvcStatusHandle, SvcStatus);
   end;
end;

procedure SvcCtrlHandlerEx(Ctrl, EventType: dword; EventData, Context: pointer); stdcall;
{ Receive a service control code from the Windows SCM, and handle it. }
begin
   case Ctrl of
   SERVICE_CONTROL_STOP,
   SERVICE_CONTROL_SHUTDOWN,
   SERVICE_CONTROL_PRESHUTDOWN:
   begin
      Out.writeln('Received service control: ' + GetServiceCodeStr(Ctrl));
      ReportSvcStatus(SERVICE_STOP_PENDING, 0, 1500);
      Stopping:= true;
   end;

   SERVICE_CONTROL_INTERROGATE:
      ReportSvcStatus(SERVICE_RUNNING, 0, 0);

   else begin
      Out.writeln('Received unknown service control: ' + IntToStr(Ctrl));
      ReportSvcStatus(SERVICE_STOPPED, 1, 0);
      Stopping:= true;
   end;
   end;
end;

procedure Service(Argc: dword; Argv: pointer); stdcall;
var ProposedExitCode: integer;
begin
   ProposedExitCode:= 0;
   Stopping:= false;
   SetLastError(0);

   if ServiceMode then
      hSvcStatusHandle:= RegisterServiceCtrlHandlerEx('MgSMTP', @SvcCtrlHandlerEx, nil);

   if ServiceMode and (hSvcStatusHandle = 0) then
      Out.writeln('RegisterServiceCtrlHandlerEx failed!'#13#10'Error code: ' + IntToStr(GetLastError))
   else begin
      SvcStatus.dwServiceType:= SERVICE_WIN32_OWN_PROCESS;
      SvcStatus.dwServiceSpecificExitCode:= 0;
      SvcStatus.dwCheckPoint:= 0;

      ReportSvcStatus(SERVICE_START_PENDING, 0, 1000);

      if ServiceMode then Out.writeln('Started in service mode.');

      Randomize;
      SetCurrentDir(ExtractFilePath(ParamStr(0)));

      if not DirectoryExists('mail')  then CreateDir('mail');
      if not DirectoryExists('spool') then CreateDir('spool');

      if FileExists('mgsmtp_server.ini') then begin

         Config:= TExtBoolINIFile.Create('mgsmtp_server.ini');

         if Config.ReadString('Server', 'Name', '') <> '' then begin
            MainServerConfig:= TMainServerConfig.Create(Config);
            Logger:=           TLogger.Create(Config);
            MailboxManager:=   TMailboxManager.Create(Config);
            RelayManager:=     TRelayManager.Create(Config);
            SpoolManager:=     TSpoolManager.Create(Config);
            PolicyManager:=    TPolicyManager.Create(Config);

            if Config.ReadBool('Spool', 'KeepProcessedEnvelopes', false)
            or Config.ReadBool('Spool', 'KeepProcessedEMails', false) then
               if not DirectoryExists('processed') then CreateDir('processed');

            Config.Free;

            AddDevComment(Logger);
            Logger.AddStdLine('Primary server name: ' + MainServerConfig.Name);
            Logger.AddStdLine('FCrDNS policy: ' + FCrDNSPolicyToStr(PolicyManager.FCrDNSPolicy));
            if MailboxManager.DomainSpecific then
               Logger.AddStdLine('Domain-specific mailbox support is enabled.');

            if not MailboxManager.Verify('postmaster') then
               ProposedExitCode:= 3;

            if SpoolManager.DeliveryThreadNumber > Length(GetAlphabetStr) then
               ProposedExitCode:= 4;

            if ProposedExitCode = 0 then begin
               SpoolManager.StartDeliveryThreads;
               StartListeners;

               ReportSvcStatus(SERVICE_RUNNING, 0, 0);
               while not Stopping do Sleep(1000);

               ReportSvcStatus(SERVICE_STOP_PENDING, 0, SpoolManager.ThreadWait * 2);

               StopListeners;
               SpoolManager.StopDeliveryThreads;
            end
            else begin
               case ProposedExitCode of
               3: Logger.AddStdLine('Error: Mandatory mailbox missing. Create a mailbox for "postmaster"!');
               4: Logger.AddStdLine('Error: The maximum allowed number of delivery threads is ' + IntToStr(Length(GetAlphabetStr)) + '.');
               end;
               ReportSvcStatus(SERVICE_STOPPED, ProposedExitCode, 0);
            end;

            ReportSvcStatus(SERVICE_STOP_PENDING, 0, 2000);

            Logger.AddStdLine('Clean shutdown.');

            PolicyManager.Free;
            SpoolManager.Free;
            RelayManager.Free;
            MailboxManager.Free;
            Logger.Free;
            MainServerConfig.Free;
            ReportSvcStatus(SERVICE_STOPPED, 0, 0);
         end
         else begin
            Config.Free;
            Out.writeln('Error: Server/Name is a mandatory configuration entry.'#13#10
               + 'Please configure the application properly, refer to the manual.');
            ReportSvcStatus(SERVICE_STOPPED, 2, 0);
         end;
      end
      else begin
         Out.writeln('Error: Missing configuration file.');
         ReportSvcStatus(SERVICE_STOPPED, 1, 0);
      end;
   end;
end;


begin
   Out.writeln('MegaBrutal''s SMTP Server, version ' + VERSION_STR + ', ' + IntToStr(PLATFORM_BITS) + ' bits');
   Out.writeln('Copyright (C) 2010-2014 MegaBrutal');
   AddDevComment(Out);
   Out.writeln;

   { TODO: Process arguments here. }
   Cmdline:= TArgumentParser.Create(CmdlineToStringArray, ArgumentPrefixes);
   WrongArgument:= Cmdline.ValidateArguments(ValidArguments);

   if WrongArgument = -1 then begin

      ServiceMode:= false;

      if Cmdline.IsPresent('?') or Cmdline.IsPresent('HELP') then begin
         Out.writeln('Supported arguments:');
         Out.writeln('/INSTALL   - registers the actual MgSMTP binary');
         Out.writeln('             as a Windows service.');
         Out.writeln('/UNINSTALL - unregisters the MgSMTP service.');
         Out.writeln('/USERMODE  - starts MgSMTP in user mode');
         Out.writeln('             (not recommended, should be only');
         Out.writeln('             used for debugging).');
         Out.writeln;
         Out.writeln('For more information on usage, see readme.txt.');
         Out.writeln('For license details, see license.txt.');
         Out.writeln;
         Out.writeln('If you can''t find any of these files, you');
         Out.writeln('don''t have the complete distribution of');
         Out.writeln('this software. In that case, download a');
         Out.writeln('proper distribution from SourceForge:');
         Out.writeln('https://sourceforge.net/projects/mgsmtp/');
      end

      else if (Cmdline.IsPresent('INSTALL') or Cmdline.IsPresent('UNINSTALL')) then begin
         { Register / unregister service. }
         hSCManager:= OpenSCManager(nil, nil, SC_MANAGER_ALL_ACCESS);
         if hSCManager <> 0 then begin
            if Cmdline.IsPresent('INSTALL') then begin
               if CreateService(hSCManager, 'MgSMTP', 'MegaBrutal''s SMTP Server (MgSMTP)', SC_MANAGER_ALL_ACCESS,
                  SERVICE_WIN32_OWN_PROCESS,
                  SERVICE_AUTO_START, SERVICE_ERROR_NORMAL,
                  PChar(ParamStr(0)),
                  nil,
                  nil,
                  nil,
                  nil,
                  nil) <> 0 then
                     Out.writeln('Service registered successfully.')
                  else
                     Out.writeln('CreateService failed!');
            end
            else if Cmdline.IsPresent('UNINSTALL') then begin
               hService:= OpenService(hSCManager, 'MgSMTP', SC_MANAGER_ALL_ACCESS);
               if hService <> 0 then begin
                  if DeleteService(hService) then
                     Out.writeln('Service unregistered successfully.')
                  else
                     Out.writeln('DeleteService failed!');
               end
               else Out.writeln('OpenService failed!');
            end;
         end
         else Out.writeln('OpenSCManager failed!');
      end

      else begin
         if Cmdline.IsPresent('USERMODE') then begin
            Out.writeln('Starting MgSMTP in user mode...');
            Service(0, nil);
         end
         else begin
            Out.writeln('Trying to contact Service Control Manager...');
            Out.writeln('(If you see this message on console, you tried to');
            Out.writeln('start up the program incorrectly. Your current');
            Out.writeln('attempt will fail, or it may hang under Wine.)');
            Out.writeln;
            ServiceMode:= true;
            if not StartServiceCtrlDispatcher(ServiceTable) then begin
               ServiceMode:= false;
               Out.writeln('Failed!');
               Out.writeln;
               Out.writeln('You need to start MgSMTP as a service,');
               Out.writeln('or supply proper arguments!');
               Out.writeln('Issue with /? for more information.');
            end;
         end;
      end

   end

   else begin

      Out.writeln('Invalid argument: ' + Cmdline.GetArgument(WrongArgument).Option + '!');

   end;
end.
