{
   Basic object-oriented network functions
   Copyright (C) 2010-2014 MegaBrutal

   This unit is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This unit is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
}

{
   Unit: Network
   This unit provides an object-oriented interface to manage TCP/IPv4
   connections.

   TTCPConnection - provides methods for sending/receiving buffers through
      the connection, and some support for text-based communication.

   TTCPRFCConnection - in addition to TTCPConnection, it provides methods
      to send/receive RFC-style commands and responses.

   TTCPListener - opens a port to listen on, and accepts incoming connections
      through it. Override its "HandleClient" method to serve connected
      clients.
}


{$MODE DELPHI}
unit Network;

interface
uses Classes, Sockets, SocketUtils, DNSResolve, NetRFC, Common;

const

   { Connection feature requests: }
   NET_TCP_BASIC        =  0;
   NET_TCP_RFCSUPPORT   =  1;

   { Default socket timeout: }
   DEF_SOCK_TIMEOUT     =  5 * 60000;     { 5 minutes. }


type

   TTCPConnection = class
      constructor Create; overload;
      constructor Create(const HostName: string; Port: word); overload;
      constructor Create(Socket: socket; const Addr: TSockAddr); overload;
      destructor Destroy; override;
   private
      FConnected: boolean;
      FSocket: socket;
      FHostIP: TIPNamePair;
      FSockTimeOut: DWord;
      sAddr: TSockAddr;
   public
      function Connect(const HostName: string; Port: word): boolean;
      procedure Disconnect;
      procedure ReverseDNSLookup;
      function VerifyFCrDNS: boolean;
      procedure SetSockTimeOut(TimeOut: DWord);
      function ReadBuffer(PtrBuffer: pointer; Len: size_t): ssize_t;
      function WriteBuffer(PtrBuffer: pointer; Len: size_t): ssize_t;
      function ReadLn(var Line: string): boolean;
      function WriteLn(const Line: string): boolean;
      property Connected: boolean read FConnected;
      property Socket: socket read FSocket;
      property HostIP: TIPNamePair read FHostIP;
      property SockTimeOut: DWord read FSockTimeOut write SetSockTimeOut;
   end;

   TTCPRFCConnection = class(TTCPConnection)
   public
      function ReadCommand(var Command: shortstring; var Prms: string): boolean;
      function ReadResponse(Response: TRFCReply): boolean;
      function SendCommand(Command: shortstring): boolean; overload;
      function SendCommand(Command: shortstring; Prms: string): boolean; overload;
      function SendResponse(Response: TRFCReply): boolean;
   end;


   TTCPAcceptHandler = procedure(Connection: TTCPConnection) of object;

   TTCPAcceptor = class(TThread)
      constructor Create(Handler: TTCPAcceptHandler; TCPConnection: TTCPConnection);
   protected
      FHandler: TTCPAcceptHandler;
      FTCPConnection: TTCPConnection;
      procedure Execute; override;
   end;

   TTCPListener = class(TThread)
      constructor Create(Port: word; FeatureRequest: word);
      {destructor Destroy; override;}
   private
      FFeatureRequest: word;
      FListenPort: word;
      FListenSocket: socket;
      sAddr: TSockAddr;
   protected
      procedure HandleClient(Connection: TTCPConnection); virtual; abstract;
      procedure Execute; override;
   public
      property ListenPort: word read FListenPort;
      function StartListen: boolean;
      procedure StopListen;
   end;



implementation


constructor TTCPConnection.Create;
{ Create an instance, but don't connect to anywhere yet. }
begin
   inherited Create;
   FConnected:= false;
   FSocket:= -1;
   FSockTimeOut:= DEF_SOCK_TIMEOUT;
end;

constructor TTCPConnection.Create(const HostName: string; Port: word);
{ Connect to the given port on the given hostname. }
begin
   inherited Create;
   Connect(HostName, Port);
end;

constructor TTCPConnection.Create(Socket: socket; const Addr: TSockAddr);
{ Use an already connected socket. }
begin
   inherited Create;
   FSocket:= Socket;
   sAddr:= Addr;
   FHostIP:= TIPNamePair.Create('', NetAddrToStr(Addr.sin_addr));
   FConnected:= true;
end;

destructor TTCPConnection.Destroy;
begin
   if FConnected then Disconnect;
   inherited Destroy;
end;


constructor TTCPAcceptor.Create(Handler: TTCPAcceptHandler; TCPConnection: TTCPConnection);
{ Start a connection handler on a distinct thread. }
begin
   FHandler:= Handler;
   FTCPConnection:= TCPConnection;
   FreeOnTerminate:= true;
   inherited Create(false);
end;


constructor TTCPListener.Create(Port: word; FeatureRequest: word);
begin
   FListenPort:= Port;
   FFeatureRequest:= FeatureRequest;
   FreeOnTerminate:= false;
   inherited Create(true);
end;


function TTCPConnection.Connect(const HostName: string; Port: word): boolean;
{ Resolves the given hostname, and tries to connect it on the given port. }
begin
   FSocket:= fpSocket(af_inet, sock_stream, 0);
   if (FSocket <> -1) then begin
      with sAddr do begin
         sin_family:= af_inet;
         sin_port:= htons(Port);
         { Resolve hostname to IP address. }
         sin_addr:= ResolveHost(HostName);
      end;

      if sAddr.sin_addr.s_addr <> 0 then
         { Try to initiate connection. }
         FConnected:= fpConnect(FSocket, @sAddr, SizeOf(sAddr)) <> -1;

      if FConnected then begin
         FHostIP:= TIPNamePair.Create(HostName, NetAddrToStr(sAddr.sin_addr));
         SetSockTimeOut(FSockTimeOut);
      end
      else
         CloseSocket(FSocket);
   end;
   Result:= FConnected;
end;

procedure TTCPConnection.Disconnect;
begin
   fpShutdown(FSocket, 2);
   CloseSocket(FSocket);
   FSocket:= -1;
   FHostIP.Free;
   FConnected:= false;
end;

procedure TTCPConnection.ReverseDNSLookup;
{ Performs a reverse DNS lookup, and updates the HostIP structure. }
var NHostIP: TIPNamePair;
begin
   if FConnected then begin
      NHostIP:= TIPNamePair.Create(ResolveIP(sAddr.sin_addr), FHostIP.IP);
      FHostIP.Free;
      FHostIP:= NHostIP;
   end;
end;

function TTCPConnection.VerifyFCrDNS: boolean;
begin
   Result:= NetAddrToStr(ResolveHost(HostIP.Name)) = HostIP.IP;
end;

procedure TTCPConnection.SetSockTimeOut(TimeOut: DWord);
begin
   FSockTimeOut:= TimeOut;
   if Connected then begin
      fpSetSockOpt(FSocket, SOL_SOCKET, SO_RCVTIMEO, @FSockTimeOut, SizeOf(FSockTimeOut));
      fpSetSockOpt(FSocket, SOL_SOCKET, SO_SNDTIMEO, @FSockTimeOut, SizeOf(FSockTimeOut));
   end;
end;

function TTCPConnection.ReadBuffer(PtrBuffer: pointer; Len: size_t): ssize_t;
begin
   Result:= fpRecv(FSocket, PtrBuffer, Len, 0);
end;

function TTCPConnection.WriteBuffer(PtrBuffer: pointer; Len: size_t): ssize_t;
begin
   Result:= fpSend(FSocket, PtrBuffer, Len, 0);
end;

function TTCPConnection.ReadLn(var Line: string): boolean;
begin
   Result:= SockReadLn(FSocket, Line);
end;

function TTCPConnection.WriteLn(const Line: string): boolean;
begin
   Result:= SockWriteLn(FSocket, Line);
end;


function TTCPRFCConnection.ReadCommand(var Command: shortstring; var Prms: string): boolean;
begin
   Result:= NetRFC.ReadCommand(FSocket, Command, Prms);
end;

function TTCPRFCConnection.ReadResponse(Response: TRFCReply): boolean;
begin
   Result:= NetRFC.ReadResponse(FSocket, Response);
end;

function TTCPRFCConnection.SendCommand(Command: shortstring): boolean;
begin
   Result:= NetRFC.SendCommand(FSocket, Command);
end;

function TTCPRFCConnection.SendCommand(Command: shortstring; Prms: string): boolean;
begin
   Result:= NetRFC.SendCommand(FSocket, Command, Prms);
end;

function TTCPRFCConnection.SendResponse(Response: TRFCReply): boolean;
begin
   Result:= NetRFC.SendResponse(FSocket, Response);
end;


procedure TTCPAcceptor.Execute;
begin
   FHandler(FTCPConnection);
end;


function TTCPListener.StartListen: boolean;
begin
   FListenSocket:= fpSocket(af_inet, sock_stream, 0);
   if FListenSocket <> -1 then begin
      with sAddr do begin
         sin_family:= af_inet;
         sin_port:= htons(FListenPort);
         sin_addr.s_addr:= 0;
      end;
      if fpBind(FListenSocket, @sAddr, sizeof(sAddr)) <> -1 then begin
         { It seems the maximum connection value isn't enforced by the
           Free Pascal library, so this 512 is a constant, dummy value. }
           { TEMPORARY SETTING OF 1 FROM 512! }
         if fpListen(FListenSocket, 512) <> -1 then begin
            Result:= true;
            Start;
         end
         else Result:= false;
      end
      else Result:= false;
   end
   else Result:= false;
end;

procedure TTCPListener.StopListen;
begin
   Terminate;
   KillThread(Handle);
end;

procedure TTCPListener.Execute;
var ClientSocket: socket; AcceptFailCount: word; Len: longint;
    TCPConnection: TTCPConnection;
begin
   { Now, accept connections. }
   AcceptFailCount:= 0;
   while not Terminated do begin
      Len:= SizeOf(sAddr);
      ClientSocket:= fpAccept(FListenSocket, @sAddr, @Len);
      if ClientSocket <> -1 then begin
         AcceptFailCount:= 0;

         { Creates the requested TTCPConnection object for the accepted
           connection. }
         case FFeatureRequest of
         NET_TCP_BASIC:
            TCPConnection:= TTCPConnection.Create(ClientSocket, sAddr);
         NET_TCP_RFCSUPPORT:
            TCPConnection:= TTCPRFCConnection.Create(ClientSocket, sAddr);
         end;

         { Then start a new thread with the connection handler. }
         TTCPAcceptor.Create(HandleClient, TCPConnection);
      end
      else begin
         Inc(AcceptFailCount);
         if AcceptFailCount >= 512 then Terminate;
      end;
   end;
end;


end.
