{
   MegaBrutal's SMTP Server (MgSMTP)
   Copyright (C) 2010-2014 MegaBrutal

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
}

{
   Unit: Policies
   Responsible for manage rights for hosts and users, it also authenticates
   users.
}


{$MODE DELPHI}
unit Policies;

interface
uses SysUtils, Classes, INIFiles, md5, CompareWild, Common, Spool;

const

   { Rights can be assigned to connections and users. }
   RIGHT_CONNECT      =   1;
   RIGHT_STORE        =   2;
   RIGHT_RELAY        =   4;

   DEFAULT_RIGHTS     = RIGHT_CONNECT;
   ALL_RIGHTS         = 255;


   { Password types: }
   PASSTYPE_PLAIN     =   0;
   PASSTYPE_MD5       =   1;
   PASSTYPE_INVALID   = 255;

   { FCrDNS policies: }
   FCRDNS_NAIVE       =   0;
   FCRDNS_AWARE       =   1;
   FCRDNS_MEAN        =   2;
   FCRDNS_STRICT      =   3;


type

   TRights = byte;
   TPassType = byte;

   { TRuleHolder administers the rule table (given in the configuration)
     for hosts or users, and it's capable of holding anything else's
     rights that may have rights at all. }

   TRuleHolder = class(TStringList)
      constructor Create(WildcardSupport: boolean);
   protected
      FWildcardSupport: boolean;
      Rights: array of TRights;
   public
      procedure AddRights(ItemName: string; AssignedRights: TRights);
      function GetRightsOf(ItemName: string): TRights;
      function GetRightsOfPair(FirstName, SecondName: string): TRights;
   end;

   { TPolicyObject holds the rights for one single object.
     And additionally, it holds the databytes limit for it. }

   TPolicyObject = class
      constructor Create(Rights: TRights; Databytes: longint);
   protected
      FRights: TRights;
      FDatabytes: longint;
      procedure Reset(Rights: TRights; Databytes: longint);
   public
      procedure Grant(Right: TRights);
      procedure Deny(Right: TRights);
      function HasRight(Right: TRights): boolean;
      function RightsStr: string;
      property Databytes: longint read FDatabytes;
   end;

   { TUserPolicies holds the data assigned for a single user, except
     rights. }

   TUserPolicies = class(TNamedObject)
      constructor Create(Name: string; Config: TINIFile);
   protected
      FAuth: boolean;
      FDatabytes: longint;
      FPassType: TPassType;
      FPassword: string;
   public
      function Authenticate(Password: string): boolean;
      property Databytes: longint read FDatabytes;
   end;

   { TPolicyManager is the interface for this unit. It sets up objects
     corresponding the configuration, creates policy objects on request,
     authenticates users, and so on. }

   TPolicyManager = class
      constructor Create(Config: TINIFile);
      destructor Destroy; override;
   protected
      FHideVersion, FReqHELO, FHosts, FUsers: boolean;
      FMaxAuthAttempts: integer;
      FFCrDNSPolicy: byte;
      HostRights, UserRights: TRuleHolder;
      UserData: array of TUserPolicies;
   public
      function AuthenticateUser(Username, Password: string; PolicyObject: TPolicyObject): boolean;
      function MakePolicyObject(Originator: TIPNamePair): TPolicyObject;
      procedure RevalidatePolicyObject(PolicyObject: TPolicyObject; Originator: TIPNamePair; EvaluateHostname, MeanMode: boolean);
      property HideVersion: boolean read FHideVersion;
      property ReqHELO: boolean read FReqHELO;
      property Hosts: boolean read FHosts;
      property Users: boolean read FUsers;
      property MaxAuthAttempts: integer read FMaxAuthAttempts;
      property FCrDNSPolicy: byte read FFCrDNSPolicy;
   end;


   function RightsToStr(Rights: TRights): string;
   function FCrDNSPolicyToStr(FCrDNSPolicy: byte): string;


var

   PolicyManager: TPolicyManager;


implementation


function StrToPassType(S: string): TPassType;
begin
   S:= UpperCase(S);
   if S = 'PLAIN' then     Result:= PASSTYPE_PLAIN
   else if S = 'MD5' then  Result:= PASSTYPE_MD5
   else                    Result:= PASSTYPE_INVALID;
end;

function StrToRights(S: string): TRights;
var SL: TStringList; i: integer; R: TRights;
begin
   S:= UpperCase(S);
   SL:= TStringList.Create;
   SL.Delimiter:= ',';
   SL.DelimitedText:= S;
   R:= DEFAULT_RIGHTS;

   for i:= 0 to SL.Count - 1 do begin

      if SL.Strings[i] = 'ALLOWSTORE' then
         R:= R or RIGHT_STORE

      else if SL.Strings[i] = 'ALLOWRELAY' then
         R:= R or RIGHT_RELAY

      else if SL.Strings[i] = 'CONNECT' then
         R:= R or RIGHT_CONNECT

      else if SL.Strings[i] = 'DENYSTORE' then
         R:= R and (ALL_RIGHTS - RIGHT_STORE)

      else if SL.Strings[i] = 'DENYRELAY' then
         R:= R and (ALL_RIGHTS - RIGHT_RELAY)

      else if SL.Strings[i] = 'DISCONNECT' then
         R:= R and (ALL_RIGHTS - RIGHT_CONNECT)

      else
         { !!! TODO: Somehow report error !!! }
         ;

   end;

   SL.Free;
   Result:= R;
end;

function RightsToStr(Rights: TRights): string;
var S: string;

   procedure AddSep;
   begin
      if S <> '' then S:= S + ', ';
   end;

begin
   S:= '';

   if (Rights and RIGHT_CONNECT) <> 0 then
      S:= 'CONNECT';

   if (Rights and RIGHT_STORE) <> 0 then begin
      AddSep; S:= S + 'STORE';
   end;

   if (Rights and RIGHT_RELAY) <> 0 then begin
      AddSep; S:= S + 'RELAY';
   end;

   if S = '' then S:= '<NONE>';
   Result:= S;
end;

function FCrDNSPolicyToStr(FCrDNSPolicy: byte): string;
begin
   case FCrDNSPolicy of
      FCRDNS_NAIVE:  Result:= 'Naive';
      FCRDNS_AWARE:  Result:= 'Aware';
      FCRDNS_MEAN:   Result:= 'Mean';
      FCRDNS_STRICT: Result:= 'Strict';
      else           Result:= 'Unknown';
   end;
end;


constructor TRuleHolder.Create(WildcardSupport: boolean);
begin
   inherited Create;
   FWildcardSupport:= WildcardSupport;
   SetLength(Rights, 0);
end;


constructor TPolicyObject.Create(Rights: TRights; Databytes: longint);
begin
   inherited Create;
   Reset(Rights, Databytes);
end;


constructor TUserPolicies.Create(Name: string; Config: TINIFile);
var Section: string;
begin
   Section:= 'Policies\Users\' + Name;
   inherited Create(Name, Config, Section);
   FAuth:=        Config.ReadBool(Section, 'Auth', true);
   FDatabytes:=   Config.ReadInteger(Section, 'Databytes', SpoolManager.Databytes);
   FPassType:=    StrToPassType(Config.ReadString(Section, 'PassType', 'plain'));
   FPassword:=    Config.ReadString(Section, 'Password', '');
end;


constructor TPolicyManager.Create(Config: TINIFile);
var Section, RS, TS: string; i: integer; SL: TStringList;
begin
   inherited Create;
   SetLength(UserData, 0);
   if MainServerConfig.Policies then begin
      Section:= 'Policies';
      FHideVersion:= Config.ReadBool(Section, 'HideVersion', false);
      FHosts:=       Config.ReadBool(Section, 'Hosts', true);
      FUsers:=       Config.ReadBool(Section, 'Users', true);
      FReqHELO:=     Config.ReadBool(Section, 'ReqHELO', false);

      FMaxAuthAttempts:= Config.ReadInteger(Section, 'MaxAuthAttempts', 0);

      HostRights:=   TRuleHolder.Create(true);
      UserRights:=   TRuleHolder.Create(false);

      { Load FCrDNSPolicy. }
      TS:= UpperCase(Config.ReadString(Section, 'FCrDNSPolicy', 'AWARE'));
      if TS = 'AWARE' then FFCrDNSPolicy:= FCRDNS_AWARE
      else if TS = 'MEAN' then FFCrDNSPolicy:= FCRDNS_MEAN
      else if TS = 'STRICT' then FFCrDNSPolicy:= FCRDNS_STRICT
      else FFCrDNSPolicy:= FCRDNS_NAIVE;

      SL:= TStringList.Create;

      { Load the rules (rights) for hosts. }
      Config.ReadSection(Section + '\Hosts', SL);
      for i:= 0 to SL.Count - 1 do begin
         RS:= Config.ReadString(Section + '\Hosts', SL.Strings[i], '');
         HostRights.AddRights(SL.Strings[i], StrToRights(RS));
      end;

      SL.Clear;

      { Load the rules (rights) and other data for users. }
      if FUsers then begin
         Config.ReadSection(Section + '\Users', SL);
         SetLength(UserData, SL.Count);
         for i:= 0 to SL.Count - 1 do begin
            RS:= Config.ReadString(Section + '\Users', SL.Strings[i], '');
            UserRights.AddRights(SL.Strings[i], StrToRights(RS));
            UserData[i]:= TUserPolicies.Create(SL.Strings[i], Config);
         end;
      end;

      { Disable user authentication, if there are no users. }
      FUsers:= Length(UserData) <> 0;

      SL.Free;
   end
   else begin
      FHosts:= false; FUsers:= false; FReqHELO:= false; FHideVersion:= false;
      FMaxAuthAttempts:= 0;
      FFCrDNSPolicy:= FCRDNS_NAIVE;
   end;
end;

destructor TPolicyManager.Destroy;
var i: integer;
begin
   HostRights.Free;
   UserRights.Free;
   for i:= 0 to Length(UserData) - 1 do
      UserData[i].Free;
   SetLength(UserData, 0);
   inherited Destroy;
end;


procedure TRuleHolder.AddRights(ItemName: string; AssignedRights: TRights);
var i: integer;
begin
   i:= Count;
   Add(ItemName);
   SetLength(Rights, i + 1);
   Rights[i]:= AssignedRights;
end;

function TRuleHolder.GetRightsOf(ItemName: string): TRights;
begin
   Result:= GetRightsOfPair(ItemName, '');
end;

function TRuleHolder.GetRightsOfPair(FirstName, SecondName: string): TRights;
{ GetRightsOfPair takes 2 names, usually a hostname and an IP, and checks
  which entry in the rule table matches for any of them, in the correct order.
  It only jumps to the next item of the list if neither of the names match
  for the actual one.
  If there are no rights associated to the names, DEFAULT_RIGHTS will be
  returned. }
var i: integer; f: boolean;
begin
   i:= 0; f:= false;
   while (i < Count) and (not f) do begin
      if FWildcardSupport then
         f:= WildComp(UpperCase(Strings[i]), UpperCase(FirstName)) or WildComp(UpperCase(Strings[i]), UpperCase(SecondName))
      else
         f:= (Strings[i] = FirstName) or (Strings[i] = SecondName);
      Inc(i);
   end;
   Dec(i);
   if f then Result:= Rights[i] else Result:= DEFAULT_RIGHTS;
end;


procedure TPolicyObject.Reset(Rights: TRights; Databytes: longint);
begin
   FRights:= Rights; FDatabytes:= Databytes;
end;

procedure TPolicyObject.Grant(Right: TRights);
begin
   FRights:= FRights or Right;
end;

procedure TPolicyObject.Deny(Right: TRights);
begin
   FRights:= FRights and (not Right);
end;

function TPolicyObject.HasRight(Right: TRights): boolean;
{ If "Right" has more than one right-flags set, the function only returns
  TRUE, when the object has all the rights. }
begin
   Result:= (FRights and Right) = Right;
end;

function TPolicyObject.RightsStr: string;
begin
   Result:= RightsToStr(FRights);
end;


function TUserPolicies.Authenticate(Password: string): boolean;
begin
   case FPassType of

   PASSTYPE_PLAIN:
      Result:= FPassword = Password;

   PASSTYPE_MD5:
      Result:= FPassword = MD5Print(MD5String(Password));

   else
      { I can't authenticate anything with an invalid password type. }
      Result:= false;

   end;

end;


function TPolicyManager.AuthenticateUser(Username, Password: string; PolicyObject: TPolicyObject): boolean;
{ Authenticates a user, returns true if successful, and updates PolicyObject
  with the new rights acquired. }
var i: integer; f: boolean;
begin
   if (MainServerConfig.Policies) and (FUsers) then begin
      i:= 0; f:= false;
      while (i < Length(UserData)) and (not f) do begin
         f:= UserData[i].IsItYourName(Username);
         Inc(i);
      end;
      Dec(i);

      if f then begin
         if UserData[i].Authenticate(Password) then begin
            PolicyObject.Reset(UserRights.GetRightsOf(Username), UserData[i].Databytes);
            Result:= true;
         end
         else Result:= false;
      end
      else Result:= false;
   end
   else Result:= false;
end;

function TPolicyManager.MakePolicyObject(Originator: TIPNamePair): TPolicyObject;
{ Make a policy object, and assign the rights corresponding to the given
  originator. If Policies are disabled by configuration, assign ALL_RIGHTS. }
begin
   if MainServerConfig.Policies then begin
      Result:= TPolicyObject.Create(
                  HostRights.GetRightsOfPair(Originator.Name, Originator.IP),
                  SpoolManager.Databytes
               );
   end
   else begin
      Result:= TPolicyObject.Create(
                  ALL_RIGHTS,
                  SpoolManager.Databytes
               );
   end;
end;

procedure TPolicyManager.RevalidatePolicyObject(PolicyObject: TPolicyObject; Originator: TIPNamePair; EvaluateHostname, MeanMode: boolean);
var RestrictiveMask: TRights;
begin
   if MeanMode then RestrictiveMask:= PolicyObject.FRights else RestrictiveMask:= ALL_RIGHTS;
   if EvaluateHostname then
      PolicyObject.Reset(HostRights.GetRightsOfPair(Originator.Name, Originator.IP) and RestrictiveMask, PolicyObject.Databytes)
   else
      PolicyObject.Reset(HostRights.GetRightsOf(Originator.IP) and RestrictiveMask, PolicyObject.Databytes);
end;


end.
