{
   MegaBrutal's SMTP Server (MgSMTP)
   Copyright (C) 2010-2015 MegaBrutal

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
}

{
   Unit: Spool
   The Spool unit is the soul of MgSMTP. It implements objects to create
   and read "spool objects" (e-mails being queued by the mail server),
   and delivery threads (the threads those actually delivers a queued
   message to a local mailbox or a remote server.
}


{$MODE DELPHI}
unit Spool;

interface
uses SysUtils, Classes, INIFiles, Common, Log, Relay, Bounce;

type

   TSpoolFilters = array of string;

   TSpoolConfig = record
      AllowExceedQuota: boolean;
      MaxReceivedHeaders: integer;
      ThreadWait: integer;
      TryCount, TryDelay: integer;
      TempFailNotifyFirst: boolean;
      TempFailNotify: integer;
      KeepProcessedEnvelopes: boolean;
      KeepProcessedEMails: boolean;
   end;


   PSpoolObject = ^TSpoolObject;
   TSpoolObject = class
      constructor Create(const Name: string; const SpoolConfig: TSpoolConfig);
      destructor Destroy; override;
   protected
      FName: string;
      FOpened: boolean;
      FEnvelope: TEnvelope;
      FOriginator: TIPNamePair;
      FEMailProperties: TEMailProperties;
      SpoolConfig: TSpoolConfig;
      SpoolData: TINIFile;
      MailFile: TStream;
      StringBuffer: TStrings;
   public
      property Name: string read FName;
      property Envelope: TEnvelope read FEnvelope;
      property EMailProperties: TEMailProperties read FEMailProperties;
      property Originator: TIPNamePair read FOriginator;
      property Opened: boolean read FOpened;
      function Open: boolean; virtual; abstract;
      procedure Close; virtual;
      procedure Discard; virtual; abstract;
      function IsActual(DelaySeconds: TUnixTimeStamp): boolean; virtual;
      function IsExpired(MaxTryCount: integer; BeforeIncrement: boolean): boolean; virtual;
      function GetAccessTime: TUnixTimeStamp; virtual;
      function GetCurrentTryCount: integer; virtual;
      procedure IncrementTryCount; virtual;
      procedure SetAccessTime(TimeStamp: TUnixTimeStamp); virtual;
      procedure SetThreadInfo(ThreadNum: integer; ThreadOSID: TThreadID); virtual;
      procedure Actualize; virtual;
      function GetMessageSize: longint; virtual;
   end;

   TSpoolObjectCreator = class(TSpoolObject)
      constructor Create(const SpoolConfig: TSpoolConfig; Databytes: longint; LineBuffer: integer; Originator: TIPNamePair);
      destructor Destroy; override;
   protected
      FDatabytes: longint;
      FDatabytesCounter: longint;
      FLineBuffer: integer;
      FOriginalMessageID: string;
      ReceivedCount: integer;
      ReceivingHeaders: boolean;
      HasDate, HasMessageID: boolean;
      WriteFail: boolean;
      procedure AddNewHeaders;
      procedure TransferEnvelope;
   public
      procedure SetDatabytes(DatabytesLimit: longint);
      function GetOriginalMessageID: string;
      function DeliverMessagePart(Line: string): boolean;
      function GetErrorCode: integer;
      function Open: boolean; override;
      procedure Close; override;
      procedure Discard; override;
      property Databytes: longint read FDatabytes write SetDatabytes;
      property OriginalMessageID: string read GetOriginalMessageID;
   end;

   TSpoolObjectReader = class(TSpoolObject)
      constructor Create(const Name: string; const SpoolConfig: TSpoolConfig);
   protected
      {LockFile: THandle;}
   public
      function GetHeaders: TStrings;
      function IsEOF: boolean;
      function MakeEnvelopes(Relay: boolean): TEnvelopeArray;
      procedure QuickSetDeliveryStatus(IsLocal: boolean; Recipient: string; Status: integer; RMsg: string);
      procedure SetDeliveryStatus(IsLocal: boolean; Envelope: TEnvelope; AddStatus: integer = 0);
      procedure ReadChunk(Strings: TStrings; Lines: integer);
      function Open: boolean; override;
      procedure Close; override;
      procedure Discard; override;
   end;


   TDeliveryThread = class(TThread)
      constructor Create(CreateSuspended: boolean; ThreadNum: integer; const SpoolConfig: TSpoolConfig; const SpoolFilters: TSpoolFilters);
      destructor Destroy; override;
   protected
      FFinished: boolean;
      FThreadNum: integer;
      SpoolConfig: TSpoolConfig;
      SpoolFilters: TSpoolFilters;
      function DeliverLocalMessage(SpoolObject: TSpoolObjectReader; MailboxPtr: pointer; ReturnPath, Recipient: string): integer;
      function DeliverRelayMessage(SpoolObject: TSpoolObjectReader; Relayer: TRelayer): boolean;
      procedure HandleFailure(SpoolObject: TSpoolObjectReader; IsLocal: boolean; FailEnvelope: TEnvelope; FailedRecipient: TRecipient; AddStatus: integer; FailMsg: string);
      procedure HandleDeliveryResults(SpoolObject: TSpoolObjectReader; IsLocal: boolean; Envelope, FailEnvelope: TEnvelope; AddStatus: integer; FailMsg: string);
      procedure CreateBounceMessage(SourceSpoolObject: TSpoolObjectReader; FailEnvelope: TEnvelope);
      function NeedSendReport(SpoolObject: TSpoolObject): boolean;
      procedure Execute; override;
   public
      procedure CallExecute;
      property Finished: boolean read FFinished;
      property ThreadNum: integer read FThreadNum;
   end;


   TSpoolManager = class
      constructor Create(Config: TINIFile);
   protected
      FDatabytes: longint;
      FLineBuffer: integer;
      SpoolConfig: TSpoolConfig;
      DeliveryThreads: array of TDeliveryThread;
   public
      function GetNumberOfDeliveryThreads: integer;
      function CreateSpoolObject(Originator: TIPNamePair): TSpoolObjectCreator;
      procedure DebugDeliveryThread;
      procedure StartDeliveryThreads;
      procedure StopDeliveryThreads;
      property Databytes: longint read FDatabytes;
      property AllowExceedQuota: boolean read SpoolConfig.AllowExceedQuota;
      property ThreadWait: integer read SpoolConfig.ThreadWait;
      property DeliveryThreadNumber: integer read GetNumberOfDeliveryThreads;
   end;


const

   { SpoolObjectCreator errors: }
   SCE_NO_ERROR         =  0;
   SCE_SIZE_EXCEEDED    =  1;
   SCE_LOOP_DETECTED    =  2;
   SCE_WRITE_FAIL       =  3;


var

   SpoolManager: TSpoolManager;


implementation
uses Mailbox;

const

   { Search attributes: }
   SEARCH_ATTR = faAnyFile - faDirectory - faVolumeID - faHidden;


constructor TSpoolObject.Create(const Name: string; const SpoolConfig: TSpoolConfig);
begin
   inherited Create;
   FName:= Name;
   FOpened:= false;
   FEnvelope:= TEnvelope.Create;
   FEMailProperties:= TEMailProperties.Create;
   Self.SpoolConfig:= SpoolConfig;
end;

destructor TSpoolObject.Destroy;
begin
   FEMailProperties.Free;
   FEnvelope.Free;
   inherited Destroy;
end;

constructor TSpoolObjectCreator.Create(const SpoolConfig: TSpoolConfig; Databytes: longint; LineBuffer: integer; Originator: TIPNamePair);
begin
   inherited Create(GenerateRandomString(16), SpoolConfig);
   FDatabytes:= Databytes;
   FLineBuffer:= LineBuffer;
   FOriginator:= Originator;
   WriteFail:= false;
end;

destructor TSpoolObjectCreator.Destroy;
begin
   FOriginator.Free;
   inherited Destroy;
end;

constructor TSpoolObjectReader.Create(const Name: string; const SpoolConfig: TSpoolConfig);
begin
   inherited Create(Name, SpoolConfig);
end;

constructor TDeliveryThread.Create(CreateSuspended: boolean; ThreadNum: integer; const SpoolConfig: TSpoolConfig; const SpoolFilters: TSpoolFilters);
begin
   FreeOnTerminate:= false;
   FFinished:= false;
   FThreadNum:= ThreadNum;
   Self.SpoolConfig:= SpoolConfig;
   Self.SpoolFilters:= SpoolFilters;
   inherited Create(CreateSuspended);
end;

destructor TDeliveryThread.Destroy;
begin
   SetLength(SpoolFilters, 0);
   inherited Destroy;
end;

constructor TSpoolManager.Create(Config: TINIFile);
begin
   inherited Create;
   FDatabytes:=         Config.ReadInteger('Spool', 'Databytes', MainServerConfig.Databytes);
   FLineBuffer:=        Config.ReadInteger('Spool', 'LineBuffer', 64);
   SetLength(DeliveryThreads, Config.ReadInteger('Spool', 'DeliveryThreads', 8));
   SpoolConfig.AllowExceedQuota:=         Config.ReadBool('Spool', 'AllowExceedQuota', false);
   SpoolConfig.MaxReceivedHeaders:=       Config.ReadInteger('Spool', 'MaxReceivedHeaders', 32);
   SpoolConfig.ThreadWait:=               Config.ReadInteger('Spool', 'ThreadWait', 250);
   SpoolConfig.TryCount:=                 Config.ReadInteger('Spool', 'TryCount', 0);
   SpoolConfig.TryDelay:=                 Config.ReadInteger('Spool', 'TryDelay', 2);
   SpoolConfig.TempFailNotifyFirst:=      Config.ReadBool('Spool', 'TempFailNotifyFirst', false);
   SpoolConfig.TempFailNotify:=           Config.ReadInteger('Spool', 'TempFailNotify', 720);
   SpoolConfig.KeepProcessedEnvelopes:=   Config.ReadBool('Spool', 'KeepProcessedEnvelopes', false);
   SpoolConfig.KeepProcessedEMails:=      Config.ReadBool('Spool', 'KeepProcessedEMails', false);
end;


procedure TSpoolObject.Close;
begin
   MailFile.Free;
   SpoolData.Free;
   FOpened:= false;
end;

function TSpoolObject.IsActual(DelaySeconds: TUnixTimeStamp): boolean;
begin
   Result:= UnixTimeStamp(Now) >= (GetAccessTime + DelaySeconds);
end;

function TSpoolObject.IsExpired(MaxTryCount: integer; BeforeIncrement: boolean): boolean;
begin
   if MaxTryCount > 0 then begin
      if BeforeIncrement then Dec(MaxTryCount);
      Result:= GetCurrentTryCount >= MaxTryCount;
   end
   else
      Result:= false;
end;

function TSpoolObject.GetAccessTime: TUnixTimeStamp;
begin
   Result:= SpoolData.ReadInteger('SpoolObject', 'AccessTime', 0);
end;

function TSpoolObject.GetCurrentTryCount: integer;
begin
   Result:= SpoolData.ReadInteger('SpoolObject', 'TryCount', 0);
end;

procedure TSpoolObject.IncrementTryCount;
begin
   SpoolData.WriteInteger('SpoolObject', 'TryCount', GetCurrentTryCount + 1);
end;

procedure TSpoolObject.SetAccessTime(TimeStamp: TUnixTimeStamp);
begin
   SpoolData.WriteInteger('SpoolObject', 'AccessTime', TimeStamp);
end;

procedure TSpoolObject.SetThreadInfo(ThreadNum: integer; ThreadOSID: TThreadID);
begin
   SpoolData.WriteString('SpoolObject', 'ThreadInfo', IntToStr(ThreadNum) + ',' + IntToStr(ThreadOSID));
end;

procedure TSpoolObject.Actualize;
begin
   IncrementTryCount;
   SetAccessTime(UnixTimeStamp(Now));
end;

function TSpoolObject.GetMessageSize: longint;
var SearchRec: TSearchRec;
begin
   if FindFirst('spool\' + FName + '.eml', SEARCH_ATTR, SearchRec) = 0 then
      Result:= SearchRec.Size
   else
      Result:= 0;
   FindClose(SearchRec);
end;


procedure TSpoolObjectCreator.AddNewHeaders;
begin
   { Add a date, if not present. }
   if not HasDate then StringBuffer.Insert(0, 'Date: ' + EMailTimeStampCorrected(Now));

   { Add Message-Id, if not present. }
   if not HasMessageID then StringBuffer.Insert(0,
      'Message-Id: <' + OriginalMessageID + '>');

   { Add Received by... }
   StringBuffer.Insert(0, 'Received: from ' + Originator.Name
      + ' ([' + Originator.IP + ']) ');
   StringBuffer.Insert(1, #9'by ' + MainServerConfig.Name + ' with SMTP (MgSMTP '
      + MainServerConfig.VersionStr + ')');
   StringBuffer.Insert(2, #9'id ' + Name + '; ' + EMailTimeStampCorrected(Now));

   { Flush it to the file. }
   try
      StringBuffer.SaveToStream(MailFile);
      StringBuffer.Clear;
   except
   end;
   ReceivingHeaders:= false;
end;

procedure TSpoolObjectCreator.TransferEnvelope;
{ Write the actual envelope structure to the data file of the spool object.
  (That's actually an INI file.) }
var i: integer; Recipient: TRecipient; Pref: string;
begin
   with SpoolData do begin
      WriteString('SpoolObject', 'ID', Name);
      WriteString('SpoolObject', 'Return-Path', Envelope.ReturnPath);
      WriteString('Originator', 'Name', Originator.Name);
      WriteString('Originator', 'IP', Originator.IP);
      for i:= 0 to Envelope.GetNumberOfRecipients - 1 do begin
         Recipient:= Envelope.GetRecipient(i);
         if MailboxManager.IsLocalAddress(Recipient.Address) then Pref:= 'Local\' else Pref:= 'Relay\';
         WriteInteger(Pref + EMailHost(Recipient.Address), EMailUserName(Recipient.Address),
            Recipient.Data);
      end;
   end;
end;

procedure TSpoolObjectCreator.SetDatabytes(DatabytesLimit: longint);
begin
   FDatabytes:= DatabytesLimit;
end;

function TSpoolObjectCreator.GetOriginalMessageID: string;
begin
   if HasMessageID then
      Result:= FOriginalMessageID
   else
      Result:= Name + '@' + MainServerConfig.Name;
end;

function TSpoolObjectCreator.GetErrorCode: integer;
begin
   if WriteFail then
      Result:= SCE_WRITE_FAIL
   else if (FDatabytesCounter > FDatabytes) and (FDatabytes <> 0) then
      Result:= SCE_SIZE_EXCEEDED
   else if ReceivedCount >= SpoolConfig.MaxReceivedHeaders then
      Result:= SCE_LOOP_DETECTED
   else
      Result:= SCE_NO_ERROR;
end;

function TSpoolObjectCreator.DeliverMessagePart(Line: string): boolean;
var Header, Value: string;
begin
   if Opened and (not WriteFail) then begin

      { If we haven't received all the headers of the e-mail, keep checking
        the incoming headers - we need to check for the existence of some
        headers, and add missing headers after all headers have arrived. }
      if ReceivingHeaders then begin
         if Length(Line) = 0 then begin
            { End of headers. }
            AddNewHeaders;
         end
         else if pos('MESSAGE-ID:', UpperCase(Line)) = 1 then begin
            HasMessageID:= true;
            SplitParameters(Line, Header, Value, ':');
            FOriginalMessageID:= CleanEMailAddress(Value);
         end
         else if pos('DATE:', UpperCase(Line)) = 1 then HasDate:= true
         else if pos('RECEIVED:', UpperCase(Line)) = 1 then Inc(ReceivedCount);
      end;

      { In any way, write the received line to the buffer, unless the databytes
        limit has been reached. }
      if (FDatabytesCounter <= FDatabytes) or (FDatabytes = 0) then begin
         StringBuffer.Add(Line);
         FDatabytesCounter:= FDatabytesCounter + Length(Line) + 2;
         { Don't flush the buffer, until all the headers received. }
         if not ReceivingHeaders and (StringBuffer.Count >= FLineBuffer) then begin
            try
               StringBuffer.SaveToStream(MailFile);
               StringBuffer.Clear;
               Result:= true;
            except
               Result:= false;
               WriteFail:= true;
            end;
         end
         else Result:= true;
      end
      else Result:= false;
   end
   else Result:= false;
end;

function TSpoolObjectCreator.Open: boolean;
begin
   try
      MailFile:= TFileStream.Create('spool\' + Name + '.eml', fmCreate);
      SpoolData:= TINIFile.Create('spool\' + Name + '.tmp');
      FOpened:= true;
      StringBuffer:= TStringList.Create;
      ReceivedCount:= 0;
      ReceivingHeaders:= true;
      HasDate:= false; HasMessageID:= false;
      FDatabytesCounter:= 0;
      TransferEnvelope;
      Result:= true;
   except
      MailFile.Free;
      SpoolData.Free;
      Result:= false;
   end;
end;

procedure TSpoolObjectCreator.Close;
begin
   SpoolData.WriteInteger('SpoolObject', 'Flags', EMailProperties.Flags);
   SpoolData.WriteInteger('SpoolObject', 'TryCount', 0);
   if ReceivingHeaders then AddNewHeaders;
   StringBuffer.SaveToStream(MailFile);
   inherited Close;
   StringBuffer.Free;
   RenameFile('spool\' + Name + '.tmp', 'spool\' + Name + '.dat');
end;

procedure TSpoolObjectCreator.Discard;
begin
   MailFile.Free;
   SpoolData.Free;
   DeleteFile('spool\' + FName + '.tmp');
   DeleteFile('spool\' + FName + '.eml');
   StringBuffer.Free;
   FOpened:= false;
end;


function TSpoolObjectReader.Open: boolean;
begin
   {LockFile:= FileCreate('spool\' + FName + '.lck', fmShareExclusive);
   if LockFile <> feInvalidHandle then begin}
   try
      MailFile:= TFileStream.Create('spool\' + Name + '.eml', fmOpenRead);
      { !!! TODO: Someday it would be nice to add a working buffer... !!! }
      { 16 KB read buffer - maybe it should be configurable. }
      {MailFile:= TReadBufStream.Create(TFileStream.Create('spool\' + Name + '.eml', fmOpenRead), 16 * 1024);}
      {(MailFile as TReadBufStream).SourceOwner:= true;}
      SpoolData:= TINIFile.Create('spool\' + Name + '.dat');
      Envelope.ReturnPath:= SpoolData.ReadString('SpoolObject', 'Return-Path', '');
      FOriginator:= TIPNamePair.Create(SpoolData.ReadString('Originator', 'Name', ''),
         SpoolData.ReadString('Originator', 'IP', ''));
      FEMailProperties.Size:= GetMessageSize;
      FEMailProperties.Flags:= SpoolData.ReadInteger('SpoolObject', 'Flags', 0);
      FOpened:= true;
      Result:= true;
   except
      MailFile.Free;
      SpoolData.Free;
      {FileClose(LockFile);}
      Result:= false;
   end;
end;

procedure TSpoolObjectReader.Close;
begin
   inherited Close;
   {FileClose(LockFile);}
   DeleteFile('spool\' + FName + '.lck');
   FOriginator.Free;
end;

procedure TSpoolObjectReader.Discard;
{ Discard should be called when the spool object is opened, and instead
  of Close! }
begin
   MailFile.Free;
   SpoolData.Free;
   if SpoolConfig.KeepProcessedEnvelopes then
      RenameFile('spool\' + FName + '.dat', 'processed\' + FName + '.dat');
   if SpoolConfig.KeepProcessedEMails then
      RenameFile('spool\' + FName + '.eml', 'processed\' + FName + '.eml');
   DeleteFile('spool\' + FName + '.dat');
   DeleteFile('spool\' + FName + '.eml');
   {FileClose(LockFile);}
   DeleteFile('spool\' + FName + '.lck');
   FOriginator.Free;
end;

function TSpoolObjectReader.GetHeaders: TStrings;
var Strings: TStrings; S: string; EH: boolean;
begin
   Strings:= TStringList.Create;
   MailFile.Seek(0, soFromBeginning);
   {repeat
      S:= ReadLineFromStream(MailFile);
      if S <> '' then Strings.Add(S);
   until (S = '') or (IsEOF);}
   EH:= false;
   while (not IsEOF) and (not EH) do begin
      S:= ReadLineFromStream(MailFile);
      if S <> '' then Strings.Add(S) else EH:= true;
   end;
   Result:= Strings;
end;

function TSpoolObjectReader.IsEOF: boolean;
begin
   Result:= (not Opened) or (MailFile.Position >= MailFile.Size);
end;

procedure TSpoolObjectReader.ReadChunk(Strings: TStrings; Lines: integer);
var S: string; C: integer;
begin
   C:= 0;
   while (not IsEOF) and (C < Lines) do begin
      S:= ReadLineFromStream(MailFile);
      Strings.Add(S);
      Inc(C);
   end;
end;

function TSpoolObjectReader.MakeEnvelopes(Relay: boolean): TEnvelopeArray;
var HostList, Usernames: TStringList; i, j, f: integer; Pref, Host: string;
    Env: TEnvelope;
begin
   if Opened then begin
      HostList:= TStringList.Create;
      SpoolData.ReadSections(HostList);
      i:= 0;
      while (i < HostList.Count) do begin
         if Relay then Pref:= 'Relay\' else Pref:= 'Local\';
         if pos(Pref, HostList.Strings[i]) <> 1 then
            HostList.Delete(i)
         else
            Inc(i);
      end;
      SetLength(Result, HostList.Count);
      Usernames:= TStringList.Create;
      f:= 0;
      for i:= 0 to HostList.Count - 1 do begin
         Usernames.Clear;
         Host:= Copy(HostList.Strings[i], 7, Length(HostList.Strings[i]) - 6);
         SpoolData.ReadSection(HostList.Strings[i], Usernames);
         if Usernames.Count > 0 then begin
            Env:= TEnvelope.Create;
            Env.SetReturnPath(SpoolData.ReadString('SpoolObject', 'Return-Path', ''));
            for j:= 0 to Usernames.Count - 1 do
               Env.AddRecipient(Usernames.Strings[j] + '@' + Host,
                  { It turned out we don't really need the status of the previous
                    attempt of delivery. It only caused confusion. }
                  {SpoolData.ReadInteger(HostList.Strings[i], Usernames.Strings[j], 0)}
                  0);
            Result[i-f]:= Env;
         end
         else begin
            { This is a faulty envelope which has no recipients, yet its INI section exists.
              Ignore it and go on. }
            SpoolData.EraseSection(HostList.Strings[i]);
            SetLength(Result, Length(Result) - 1);
            Inc(f);
         end;
      end;
      Usernames.Free;
      HostList.Free;
   end
   else SetLength(Result, 0);
end;

procedure TSpoolObjectReader.QuickSetDeliveryStatus(IsLocal: boolean; Recipient: string; Status: integer; RMsg: string);
{ "Quick" because it bypasses the TEnvelope structures cached in memory, so
  it writes the data immediately into the spool data file. }
var Pref, StatStr: string;
begin
   if IsLocal then Pref:= 'Local\' else Pref:= 'Relay\';
   if (Status and (DS_DELIVERED or DS_PERMANENT)) <> 0 then begin
      SpoolData.DeleteKey(Pref + EMailHost(Recipient), EMailUserName(Recipient));
      if (Status and DS_DELIVERED) <> 0 then
         StatStr:= 'Delivered'
      else
         StatStr:= 'Failed';
      Pref:= StatStr + Pref;
      Logger.AddLine('Object ' + Name, 'Permanent status has been set on recipient <' + Recipient + '>: '
         + Pref + StatusToStr(Status) + ' (' + CleanEOLN(RMsg) + ')');
   end;
   SpoolData.WriteInteger(Pref + EMailHost(Recipient), EMailUserName(Recipient), Status);
end;

procedure TSpoolObjectReader.SetDeliveryStatus(IsLocal: boolean; Envelope: TEnvelope; AddStatus: integer = 0);
{ It writes all data of an envelope to the spool data file. }
var i: integer; Recipient: TRecipient;
begin
   for i:= 0 to Envelope.GetNumberOfRecipients - 1 do begin
      Recipient:= Envelope.GetRecipient(i);
      QuickSetDeliveryStatus(IsLocal, Recipient.Address, Recipient.Data or AddStatus, Recipient.RMsg);
   end;
end;


function TDeliveryThread.DeliverLocalMessage(SpoolObject: TSpoolObjectReader; MailboxPtr: pointer; ReturnPath, Recipient: string): integer;
var LockID: integer; Headers, Chunk: TStrings; R: boolean;
    { This absolute declaration is necessary to avoid circular unit depedency
      between Spool and Mailbox, ever since Mailbox creates spool objects
      to implement forwarding/remailing. }
    Mailbox: PMailbox absolute MailboxPtr;
begin
   { !!! TODO: Change return values to named constants !!! }
   LockID:= Mailbox^.Lock;
   if LockID <> 0 then begin
      Headers:= SpoolObject.GetHeaders;
      if Mailbox^.BeginDeliverMessage(LockID, ReturnPath, Recipient, SpoolObject.Name, SpoolObject.EMailProperties, Headers) then begin
         Chunk:= TStringList.Create;
         R:= true;
         while (not SpoolObject.IsEOF) and R do begin
            { Maybe constant "32" should be configurable? }
            Chunk.Clear;
            SpoolObject.ReadChunk(Chunk, 32);
            R:= Mailbox^.DeliverMessagePart(LockID, Chunk);
         end;
         if R then begin
            if Mailbox^.FinishDeliverMessage(LockID) then
               Result:= 0
            else
               Result:= 4;
         end
         else
            Result:= 3;
         Chunk.Free;
      end
      else Result:= 2;
      Headers.Free;
      Mailbox^.Release(LockID);
   end
   else
      Result:= 1;
end;

function TDeliveryThread.DeliverRelayMessage(SpoolObject: TSpoolObjectReader; Relayer: TRelayer): boolean;
{ Relay message to remote server.
  Returns true if the transaction went through (doesn't necessarily mean
  that the message was actually accepted). }
var Headers, Chunk: TStrings; R: boolean;
begin
   if Relayer.PrepareSendMessage then begin
      Headers:= SpoolObject.GetHeaders;
      Chunk:= TStringList.Create;
      { Leave a line between the headers and the body. }
      Headers.Add('');

      R:= Relayer.DeliverMessagePart(Headers);
      while (not SpoolObject.IsEOF) and R do begin
         { Maybe constant "64" should be configurable? }
         Chunk.Clear;
         SpoolObject.ReadChunk(Chunk, 64);
         R:= Relayer.DeliverMessagePart(Chunk);
      end;
      if R then begin
         Relayer.FinishDeliverMessage;
      end
      else
         SpoolObject.SetDeliveryStatus(false, Relayer.Envelope, DS_DELAYED or DS_CONNECTIONFAIL);

      Result:= R;
      Chunk.Free;
      Headers.Free;
   end
   else Result:= false;
end;

function TDeliveryThread.NeedSendReport(SpoolObject: TSpoolObject): boolean;
{ Check if there is necessary to send a temporary failure notification,
  according to the configuration. }
var CurrentTryCount: integer;
begin
   CurrentTryCount:= SpoolObject.GetCurrentTryCount;
   Result:= ((CurrentTryCount = 0) and SpoolConfig.TempFailNotifyFirst)
      or ((CurrentTryCount <> 0) and ((CurrentTryCount mod SpoolConfig.TempFailNotify) = 0));
end;

procedure TDeliveryThread.HandleFailure(SpoolObject: TSpoolObjectReader; IsLocal: boolean; FailEnvelope: TEnvelope; FailedRecipient: TRecipient; AddStatus: integer; FailMsg: string);
{ Administer failure on a single recipient. }
begin
   if Length(FailMsg) <> 0 then FailedRecipient.RMsg:= FailMsg;
   FailedRecipient.Data:= FailedRecipient.Data or AddStatus;
   {CreateBounceMessage(SpoolObject, FailedRecipient, ReturnPath, FailMsg);}
   FailEnvelope.AddRecipient(FailedRecipient);
   SpoolObject.QuickSetDeliveryStatus(IsLocal, FailedRecipient.Address, FailedRecipient.Data, FailedRecipient.RMsg);
end;

procedure TDeliveryThread.HandleDeliveryResults(SpoolObject: TSpoolObjectReader; IsLocal: boolean; Envelope, FailEnvelope: TEnvelope; AddStatus: integer; FailMsg: string);
{ Administer results on multiple recipients (passed in a TEnvelope). }
var i: integer; Recipient: TRecipient; Expired: boolean;
begin
   Expired:= SpoolObject.IsExpired(SpoolConfig.TryCount, true);
   for i:= 0 to Envelope.GetNumberOfRecipients - 1 do begin
      Recipient:= Envelope.GetRecipient(i);
      Recipient.Data:= Recipient.Data or AddStatus;
      if Expired then
         Recipient.Data:= (Recipient.Data or DS_PERMANENT) and (DS_ALLFLAGS xor DS_DELAYED);
      if (Recipient.Data and DS_DELIVERED) <> 0 then
         SpoolObject.QuickSetDeliveryStatus(IsLocal, Recipient.Address, Recipient.Data, Recipient.RMsg)
      else begin
         if ((Recipient.Data and DS_PERMANENT) <> 0)
         or (((Recipient.Data and DS_DELAYED) <> 0) and NeedSendReport(SpoolObject)) then begin
            { In the case of failures, HandleFailure will call QuickSetDeliveryStatus. }
            HandleFailure(SpoolObject, IsLocal, FailEnvelope, Recipient, 0, FailMsg);
         end
         else
            SpoolObject.QuickSetDeliveryStatus(IsLocal, Recipient.Address, Recipient.Data, Recipient.RMsg);
      end;
   end;
end;

procedure TDeliveryThread.CreateBounceMessage(SourceSpoolObject: TSpoolObjectReader; FailEnvelope: TEnvelope);
{ Generates failure notification messages, and places them into a new spool
  object to queue them for delivery. }
var BounceSpoolObject: TSpoolObjectCreator; Headers, BounceMessage: TStrings; i: integer;
    FailedRecipient: TRecipient;
begin
   { Don't do anything, if we don't have a return-path. }
   if (FailEnvelope.ReturnPath <> '') and (FailEnvelope.GetNumberOfRecipients <> 0) then begin
      BounceSpoolObject:= TSpoolObjectCreator.Create(SpoolConfig, 1024 * 1024, 32, TIPNamePair.Create('localhost', '127.0.0.1'));
      BounceSpoolObject.Envelope.SetReturnPath('');
      BounceSpoolObject.Envelope.AddRecipient(FailEnvelope.ReturnPath);
      if BounceSpoolObject.Open then begin
         Headers:= SourceSpoolObject.GetHeaders;

         if FailEnvelope.GetNumberOfRecipients = 1 then
            BounceMessage:= GenerateBounceMessage(FailEnvelope.GetRecipient(0), Headers, FailEnvelope.ReturnPath)
         else
            BounceMessage:= GenerateBounceMessage(FailEnvelope, Headers);

         for i:= 0 to BounceMessage.Count - 1 do
            BounceSpoolObject.DeliverMessagePart(BounceMessage.Strings[i]);

         BounceSpoolObject.Close;

         for i:= 0 to FailEnvelope.GetNumberOfRecipients - 1 do begin
            FailedRecipient:= FailEnvelope.GetRecipient(i);
            Logger.AddLine('Spool', 'Bounce message created in ' + BounceSpoolObject.Name
               + ' for object ' + SourceSpoolObject.Name
               + ' for address <' + FailEnvelope.ReturnPath
               + '>; concerning recipient <' + FailedRecipient.Address
               + '>; reported status: ' + StatusToStr(FailedRecipient.Data) + ' (' + CleanEOLN(FailedRecipient.RMsg) + ')');
         end;

         BounceMessage.Free;
         Headers.Free;
      end;
      BounceSpoolObject.Free;
   end;
end;

procedure TDeliveryThread.Execute;
{ This is a very important thread, because this delivers e-mails to local
  mailboxes and to remote servers. }
var SearchRec: TSearchRec; SR: longint; SpoolObject: TSpoolObjectReader;
    Found: boolean; Envelopes: TEnvelopeArray;
    CurrEnv, FailEnv: TEnvelope; CurrRec: TRecipient;
    Mailbox: PMailbox; Relayer: TRelayer;
    NumOfEnvelopes: integer;
    a, i, j, r: integer;
begin
   while not Terminated do begin
      for a:= 0 to Length(SpoolFilters) - 1 do begin
         if FindFirst('spool\' + SpoolFilters[a] + '*.dat', SEARCH_ATTR, SearchRec) = 0 then begin
            repeat
               Found:= false; SR:= 0;
               { Try to find a spool object that's not busy, and also actual. }
               repeat
                  SpoolObject:= TSpoolObjectReader.Create(Copy(SearchRec.Name, 1, Length(SearchRec.Name) - 4), SpoolConfig);
                  if not SpoolObject.Open then begin
                     SpoolObject.Free;
                     SR:= FindNext(SearchRec);
                  end
                  else if not SpoolObject.IsActual(SpoolConfig.TryDelay * 60) then begin
                     SpoolObject.Close;
                     SpoolObject.Free;
                     SR:= FindNext(SearchRec);
                  end
                  else Found:= true;
               until Found or (SR <> 0);
               if Found then begin
                  NumOfEnvelopes:= -1;

                  FailEnv:= TEnvelope.Create;
                  FailEnv.ReturnPath:= SpoolObject.Envelope.ReturnPath;

                  { Check local addresses first. }
                  Envelopes:= SpoolObject.MakeEnvelopes(false);
                  NumOfEnvelopes:= Length(Envelopes);
                  for i:= 0 to Length(Envelopes) - 1 do begin
                     CurrEnv:= Envelopes[i];
                     for j:= 0 to CurrEnv.GetNumberOfRecipients - 1 do begin
                        CurrRec:= CurrEnv.GetRecipient(j);
                        Mailbox:= MailboxManager.GetMailbox(EMailUserName(CurrRec.Address), EMailHost(CurrRec.Address));
                        if Mailbox <> nil then begin
                           if SpoolConfig.AllowExceedQuota or Mailbox^.CheckQuota(SpoolObject.GetMessageSize) then begin
                              r:= DeliverLocalMessage(SpoolObject, Mailbox, CurrEnv.ReturnPath, CurrRec.Address);
                              if r > 1 then
                                 HandleFailure(SpoolObject, true, FailEnv, CurrRec, DS_PERMANENT or DS_INTERNALFAIL,
                                    DSMSG_INTERNALFAIL + 'DeliverLocalMessage = ' + IntToStr(r))
                              else if r = 0 then
                                 SpoolObject.QuickSetDeliveryStatus(true, CurrRec.Address, DS_DELIVERED, CurrRec.RMsg)
                              else
                                 SpoolObject.QuickSetDeliveryStatus(true, CurrRec.Address, r, CurrRec.RMsg);
                           end
                           else
                              HandleFailure(SpoolObject, true, FailEnv, CurrRec, DS_PERMANENT, DSMSG_QUOTAEXCEEDED);
                        end
                        else
                           HandleFailure(SpoolObject, true, FailEnv, CurrRec, DS_PERMANENT, DSMSG_MAILBOXNOTEXISTS);
                     end;
                     { Free envelope. }
                     CurrEnv.Free;
                  end;

                  { Check relay addresses as well. }
                  SetLength(Envelopes, 0);
                  Envelopes:= RelayManager.OrganizeEnvelopes(SpoolObject.MakeEnvelopes(true));
                  NumOfEnvelopes:= NumOfEnvelopes + Length(Envelopes);
                  for i:= 0 to Length(Envelopes) - 1 do begin
                     CurrEnv:= Envelopes[i];
                     Relayer:= RelayManager.CreateRelayer(CurrEnv, SpoolObject.EMailProperties);
                     if Relayer.OpenConnection then begin
                        if Relayer.Greet then
                           if Relayer.SendEnvelope then
                              DeliverRelayMessage(SpoolObject, Relayer);

                        AssignDeliveryStatusToSMTPCodes(CurrEnv, Relayer.IsTransactionComplete);
                        Relayer.CloseConnection;
                        Relayer.Free;
                        HandleDeliveryResults(SpoolObject, false, CurrEnv, FailEnv, 0, '');
                     end
                     else begin
                        HandleDeliveryResults(SpoolObject, false, CurrEnv, FailEnv, DS_DELAYED or DS_CONNECTIONFAIL, DSMSG_CONNECTIONFAIL + Relayer.RelayServerName);
                     end;
                     { Free envelope. }
                     CurrEnv.Free;
                  end;

                  { Create a bounce message if necessary. }
                  CreateBounceMessage(SpoolObject, FailEnv);
                  FailEnv.Free;

                  SpoolObject.Actualize;
                  SpoolObject.SetThreadInfo(ThreadNum, ThreadID);

                  if (NumOfEnvelopes <> 0) and (not SpoolObject.IsExpired(SpoolConfig.TryCount, false)) then
                     SpoolObject.Close
                  else begin
                     SpoolObject.Discard;
                     Logger.AddLine('Spool', 'Object ' + SpoolObject.Name + ' has been processed.');
                  end;
                  SpoolObject.Free;
               end;
            until (SR <> 0) or (FindNext(SearchRec) <> 0);
         end;
         FindClose(SearchRec);
      end;
      Sleep(SpoolConfig.ThreadWait);
   end;
   FFinished:= true;
end;

procedure TDeliveryThread.CallExecute;
begin
   Execute;
end;


function TSpoolManager.GetNumberOfDeliveryThreads: integer;
begin
   Result:= Length(DeliveryThreads);
end;

function TSpoolManager.CreateSpoolObject(Originator: TIPNamePair): TSpoolObjectCreator;
begin
   Result:= TSpoolObjectCreator.Create(SpoolConfig, FDatabytes, FLineBuffer, Originator);
end;

procedure TSpoolManager.DebugDeliveryThread;
{ You only need it when you need to trace the delivery thread.
  Normally it never gets called. Write a separate program to use it.
  (I've presented one, test_threaddebug.pas.) }
var i: integer; Delivery: TDeliveryThread; SpoolFilters: TSpoolFilters;
    Alphabet: string;
begin
   Alphabet:= GetAlphabetStr;
   SetLength(SpoolFilters, Length(Alphabet));
   for i:= 1 to Length(Alphabet) do SpoolFilters[i - 1]:= Alphabet[i];

   Delivery:= TDeliveryThread.Create(true, 0, SpoolConfig, SpoolFilters);
   Delivery.CallExecute;
   Delivery.Free;
end;

procedure TSpoolManager.StartDeliveryThreads;
var i, j, n, x: integer; ThreadFilters: array of TSpoolFilters; Alphabet: string;
begin
   n:= Length(DeliveryThreads);
   SetLength(ThreadFilters, n);
   Alphabet:= GetAlphabetStr;

   if n > 0 then begin
      for i:= 1 to Length(Alphabet) do begin
         x:= (i - 1) mod n;
         j:= Length(ThreadFilters[x]);
         SetLength(ThreadFilters[x], j + 1);
         ThreadFilters[x][j]:= Alphabet[i];
      end;
   end;

   Logger.AddStdLine('Spool', 'Starting ' + IntToStr(n) + ' delivery threads...');
   for i:= 0 to n - 1 do begin
      DeliveryThreads[i]:= TDeliveryThread.Create(false, i, SpoolConfig, ThreadFilters[i]);
      Sleep(25);
   end;
   Logger.AddStdLine('Spool', 'Delivery threads have been started.');
end;

procedure TSpoolManager.StopDeliveryThreads;
{ Signals delivery threads to end, and waits for them to quit. }
var i, Counter: integer; AllFinished: boolean;
begin
   Logger.AddStdLine('Spool', 'Stopping delivery threads...');
   for i:= 0 to Length(DeliveryThreads) - 1 do
      DeliveryThreads[i].Terminate;

   Counter:= 0;

   repeat
      Sleep(50);
      AllFinished:= true;
      for i:= 0 to Length(DeliveryThreads) - 1 do
         if not DeliveryThreads[i].Finished then AllFinished:= false;
      Inc(Counter);
   until AllFinished or (Counter >= 600);

   { Threads those didn't finish on time will be terminated. }
   for i:= 0 to Length(DeliveryThreads) - 1 do begin
      if not DeliveryThreads[i].Finished then begin
         Logger.AddStdLine('Spool', 'WARNING: Delivery thread #' + IntToStr(i) + ' hasn''t finished properly on time!');
         //DeliveryThreads[i].Suspend;    { Suspend has been deprecated, but we'll kill the thread regardless. }
         KillThread(DeliveryThreads[i].Handle);
      end;
      DeliveryThreads[i].Free;
   end;
   Logger.AddStdLine('Spool', 'Delivery threads have been stopped.');
end;


end.
